/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework;

import static org.ow2.petals.component.framework.Constants.Component.ACCEPTOR_SIZE;
import static org.ow2.petals.component.framework.Constants.Component.ACKNOWLEDGEMENT_TIMEOUT;
import static org.ow2.petals.component.framework.Constants.Component.DEFAULT_ACCEPTOR_SIZE;
import static org.ow2.petals.component.framework.Constants.Component.DEFAULT_ACKNOWLEDGEMENT_TIMEOUT;
import static org.ow2.petals.component.framework.Constants.Component.DEFAULT_IGNORED_STATUS;
import static org.ow2.petals.component.framework.Constants.Component.DEFAULT_POOL_SIZE;
import static org.ow2.petals.component.framework.Constants.Component.EXTERNAL_LISTENER_CLASSNAME;
import static org.ow2.petals.component.framework.Constants.Component.IGNORED_STATUS;
import static org.ow2.petals.component.framework.Constants.Component.JBI_LISTENER_CLASSNAME;
import static org.ow2.petals.component.framework.Constants.Component.PLACEHOLDER_PROPERTIES_FILE;
import static org.ow2.petals.component.framework.Constants.Component.POOL_SIZE;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;

import org.ow2.petals.component.framework.Constants.IgnoredStatusConstants;
import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.interceptor.Interceptor;
import org.ow2.petals.component.framework.interceptor.description.InterceptorBuilder;
import org.ow2.petals.component.framework.interceptor.description.InterceptorConfiguration;
import org.ow2.petals.component.framework.listener.AsyncMessageHandler;
import org.ow2.petals.component.framework.listener.JBIAcceptorManager;
import org.ow2.petals.component.framework.listener.JBIWorkerManager;
import org.ow2.petals.component.framework.listener.WorkerPoolManager;
import org.ow2.petals.component.framework.mbean.ComponentInformationMBean;
import org.ow2.petals.component.framework.mbean.MBeanHelper;
import org.ow2.petals.component.framework.mbean.PerformanceNotification;
import org.ow2.petals.component.framework.mbean.PerformanceNotifierMBean;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationMBean;
import org.ow2.petals.component.framework.mbean.RuntimeConfiguration;
import org.ow2.petals.component.framework.mbean.InterceptorManager;
import org.ow2.petals.component.framework.su.ServiceUnitExtensionsValidator;
import org.ow2.petals.component.framework.su.ServiceUnitListener;
import org.ow2.petals.component.framework.su.ServiceUnitManager;
import org.ow2.petals.component.framework.util.ComponentPropertiesManager;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.component.framework.util.PropertiesUtil;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.jbi.descriptor.Consumes;
import org.ow2.petals.jbi.descriptor.Provides;
import org.ow2.petals.jbi.descriptor.ServiceUnitExtensibleElement;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

/**
 * An abstract class that can be used to easily create a new JBI Component. It
 * implements Component, ComponentLifeCycle and MessageExchangeProcessor JBI
 * interfaces.
 * 
 * @author ofabre
 * 
 */
public abstract class AbstractComponent implements Component, ComponentLifeCycle {

    /**
     * The component's delivery channel
     */
    private DeliveryChannel channel;

    /**
     * The Component's Context
     */
    protected ComponentContext context;

    /**
     * The component configuration
     */
    private ComponentConfiguration componentConfiguration;

    /**
     * The message exchange acceptor manager
     */
    private JBIAcceptorManager acceptorManager;

    /**
     * The component's logger
     */
    private Logger log;

    /**
     * The ServiceUnit lifecycle listener
     */
    private ServiceUnitListener serviceUnitListener;

    /**
     * The service unit extensions validator
     */
    private ServiceUnitExtensionsValidator serviceUnitExtensionsValidator;

    /**
     * The component's service unit manager
     */
    private ServiceUnitManager suManager;

    /**
     * The Worker manager in charge of handling JBI Message Exchanges
     */
    private WorkerPoolManager workerManager;

    /**
     * All the component interceptors. Key is the name of the interceptor
     */
    protected Map<String, Interceptor> interceptors;

    /**
     * The asynchronous message handler
     */
    private AsyncMessageHandler asyncMessageHandler;

    /**
     * JMX Performance notifier MBean
     */
    private PerformanceNotifierMBean jmxPerformanceNotifierMBean = null;

    /**
     * JMX Component Configuration Information MBean
     */
    private ComponentInformationMBean componentInformationMBean = null;

    /**
     * The runtime configuration
     */
    private RuntimeConfiguration runtimeConfiguration = null;

    /**
     * Manage the interceptors at runtime
     */
    private InterceptorManager interceptorManager = null;

    /**
     * Creates a new {@link JBIAcceptorManager}
     * 
     * @return the created {@link JBIAcceptorManager}
     */
    private JBIAcceptorManager createJBIAcceptorManager() {
        return new JBIAcceptorManager(this, this.componentConfiguration, this.workerManager,
                this.runtimeConfiguration, this.log);
    }

    /**
     * Create the worker manager
     * 
     * @return
     */
    protected WorkerPoolManager createWorkerManager() {
        return new JBIWorkerManager(this, this.componentConfiguration, this.runtimeConfiguration,
                this.log);
    }

    /**
     * Can be overridden to provide a specific service unit manager
     * 
     * @return the service unit manager used by this component.
     */
    protected ServiceUnitManager createServiceUnitManager() {
        return new ServiceUnitManager(this);
    }

    /**
     * Creates the component configuration from the component manager. The
     * component configuration contains all CDK properties as attributes.
     * 
     * @return the component configuration.
     */
    protected ComponentConfiguration createComponentConfiguration() throws PEtALSCDKException {
        final ComponentConfiguration conf = new ComponentConfiguration();

        // get the component properties manager to load extension values
        ComponentPropertiesManager propertiesManager = new ComponentPropertiesManager();
        propertiesManager.setRootPath(this.getContext().getInstallRoot());

        if (!propertiesManager.load()) {
            throw new PEtALSCDKException(
                    "Failed to load extensions section in the component descriptor file");
        }

        final Properties properties = propertiesManager.getProperties();

        // Get properties file name (placeholders)
        Object placeholder = properties.get(PLACEHOLDER_PROPERTIES_FILE);
        if (placeholder != null) {
            conf.setPropertiesFile(placeholder.toString());
        }

        // Get pool size
        int poolSize = 0;
        try {
            if ((propertiesManager.getProperty(POOL_SIZE) == null)
                    || (propertiesManager.getProperty(POOL_SIZE).trim().equals(""))) {
                propertiesManager.getProperties().setProperty(POOL_SIZE,
                        Integer.toString(DEFAULT_POOL_SIZE));
            }
            poolSize = Integer.parseInt(propertiesManager.getProperty(POOL_SIZE, Integer
                    .toString(DEFAULT_POOL_SIZE)));
        } catch (final NumberFormatException e) {
            throw new PEtALSCDKException("The CDK parameter '" + POOL_SIZE
                    + "' is invalid in the component descriptor file");
        }
        if (poolSize <= 0) {
            throw new PEtALSCDKException("The CDK parameter '" + POOL_SIZE
                    + "' is invalid in the component descriptor file");
        }
        conf.setPoolSize(poolSize);
        properties.setProperty(POOL_SIZE, Integer.toString(conf.getPoolSize()));

        // Get the JBI listener acceptor size
        int acceptorSize = 0;
        try {
            if ((propertiesManager.getProperty(ACCEPTOR_SIZE) == null)
                    || (propertiesManager.getProperty(ACCEPTOR_SIZE).trim().equals(""))) {
                propertiesManager.getProperties().setProperty(ACCEPTOR_SIZE,
                        Integer.toString(DEFAULT_ACCEPTOR_SIZE));
            }
            acceptorSize = Integer.parseInt(propertiesManager.getProperty(ACCEPTOR_SIZE, Integer
                    .toString(DEFAULT_ACCEPTOR_SIZE)));
        } catch (final NumberFormatException e) {
            throw new PEtALSCDKException("The CDK parameter '" + ACCEPTOR_SIZE
                    + "' is invalid in the component descriptor file (Not a Number)");
        }
        if (acceptorSize <= 0) {
            throw new PEtALSCDKException("The CDK parameter '" + ACCEPTOR_SIZE
                    + "' is invalid in the component descriptor file (<0)");
        }
        conf.setAcceptorSize(acceptorSize);
        properties.setProperty(ACCEPTOR_SIZE, Integer.toString(conf.getAcceptorSize()));

        // Get ignored status
        IgnoredStatusConstants ignoreStatus;
        try {
            if ((propertiesManager.getProperty(IGNORED_STATUS) == null)
                    || (propertiesManager.getProperty(IGNORED_STATUS).trim().equals(""))) {
                propertiesManager.getProperties().setProperty(IGNORED_STATUS,
                        DEFAULT_IGNORED_STATUS);
            }
            ignoreStatus = IgnoredStatusConstants.valueOf(propertiesManager.getProperty(
                    IGNORED_STATUS, DEFAULT_IGNORED_STATUS));

        } catch (final IllegalArgumentException e) {
            throw new PEtALSCDKException("The CDK parameter '" + IGNORED_STATUS
                    + "' is invalid in the component descriptor file");
        }
        conf.setIgnoredStatus(ignoreStatus);
        properties.setProperty(IGNORED_STATUS, conf.getIgnoredStatus().value());

        // Get the ack timeout
        int ackTimeout = 0;
        try {
            if ((propertiesManager.getProperty(ACKNOWLEDGEMENT_TIMEOUT) == null)
                    || (propertiesManager.getProperty(ACKNOWLEDGEMENT_TIMEOUT).trim().equals(""))) {
                propertiesManager.getProperties().setProperty(ACKNOWLEDGEMENT_TIMEOUT,
                        Long.toString(DEFAULT_ACKNOWLEDGEMENT_TIMEOUT));
            }
            ackTimeout = Integer.parseInt(propertiesManager.getProperty(ACKNOWLEDGEMENT_TIMEOUT,
                    Long.toString(DEFAULT_ACKNOWLEDGEMENT_TIMEOUT)));
        } catch (final NumberFormatException e) {
            throw new PEtALSCDKException("The CDK parameter '" + ACKNOWLEDGEMENT_TIMEOUT
                    + "' is invalid in the component descriptor file");
        }
        if (ackTimeout <= 0) {
            throw new PEtALSCDKException("The CDK parameter '" + ACKNOWLEDGEMENT_TIMEOUT
                    + "' is invalid in the component descriptor file");
        }
        conf.setAckTimeout(ackTimeout);
        properties.setProperty(ACKNOWLEDGEMENT_TIMEOUT, Long.toString(conf.getAckTimeout()));

        // Get the JBI listener, all the CDK based components MUST have a JBI
        // listener
        final String jbiListenerClassName = propertiesManager.getProperty(JBI_LISTENER_CLASSNAME);
        if (StringHelper.isNullOrEmpty(jbiListenerClassName)) {
            throw new PEtALSCDKException("The CDK parameter '" + JBI_LISTENER_CLASSNAME
                    + "' is invalid or absent in the component descriptor file");
        }
        conf.setJbiListenerClassName(jbiListenerClassName);

        // Get the external listener (optional)
        conf.setExternalListenerClassName(propertiesManager
                .getProperty(EXTERNAL_LISTENER_CLASSNAME));

        this.log.log(Level.INFO, "Component Framework configuration :");
        this.log.log(Level.INFO, " - JBI Acceptors size : " + conf.getAcceptorSize());
        this.log.log(Level.INFO, " - JBI Workers size : " + conf.getPoolSize());
        this.log.log(Level.INFO, " - Ignored Message Status : " + conf.getIgnoredStatus().value());
        this.log.log(Level.INFO, " - Ack TimeOut : " + conf.getAckTimeout() + " ms");
        this.log.log(Level.INFO, " - JBI Listener : " + conf.getJbiListenerClassName());
        this.log.log(Level.INFO, " - External Listener : " + conf.getExternalListenerClassName());

        // put only the updated properties in the component configuration
        this.log.info("Component parameters :");
        Set<Object> set = properties.keySet();
        Iterator<Object> iter = set.iterator();
        Map<String, String> map = new HashMap<String, String>();
        while (iter.hasNext()) {
            String name = (String) iter.next();
            if (name.startsWith(Constants.PARAMETER_PREFIX)) {
                String paramName = name.substring(Constants.PARAMETER_PREFIX.length(), name
                        .length());
                String value = (String) properties.get(paramName);
                if (value == null) {
                    // avoid NPE...
                    value = "";
                }
                this.log.info(" - " + paramName + " : '" + value + "'");
                map.put(paramName, value);
            }
        }
        // clear and save
        conf.getProperties().clear();
        conf.getProperties().putAll(map);

        return conf;
    }

    /**
     * Returns the component delivery channel, used to send messages to petals
     * container
     * 
     * @return the delivery channel. Not null.
     */
    public DeliveryChannel getChannel() {
        return this.channel;
    }

    /**
     * The component context
     * 
     * @return the component context. Not null.
     */
    public ComponentContext getContext() {
        return this.context;
    }

    /**
     * Returns null by default as this class does not provides JMX extension
     * MBean
     * 
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#getLifeCycle()
     */
    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    /**
     * Returns the component logger
     * 
     * @return component logger. Not null.
     */
    public Logger getLogger() {
        return this.log;
    }

    /**
     * Can be overidden to return a custom resource bundle name
     * 
     * @return the resource bundle name or null if no resource bundle name has
     *         to be used (default value)
     */
    protected String getResourceBundleName() {
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#getServiceDescription(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public Document getServiceDescription(final ServiceEndpoint endpoint) {
        return this.suManager.getServiceDescription(endpoint);
    }

    public ServiceUnitListener getServiceUnitListener() {
        return this.serviceUnitListener;
    }

    public ServiceUnitExtensionsValidator getServiceUnitExtensionsValidator() {
        return this.serviceUnitExtensionsValidator;
    }

    /**
     * Get all the interceptors
     * 
     * @return
     */
    public Map<String, Interceptor> getInterceptors() {
        return this.interceptors;
    }

    /**
     * Get the interceptor that have the given name.
     * 
     * @param name
     * @return null if the interceptor is not found of is there are no
     *         interceptor in the component
     */
    public Interceptor getInterceptor(final String name) {
        Interceptor interceptor = null;

        if (this.interceptors != null) {
            interceptor = this.interceptors.get(name);
        }
        return interceptor;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#getServiceUnitManager()
     */
    public javax.jbi.component.ServiceUnitManager getServiceUnitManager() {
        if (this.suManager == null) {
            this.suManager = this.createServiceUnitManager();
        }
        return this.suManager;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#init(javax.jbi.component.ComponentContext)
     */
    public final void init(final ComponentContext compContext) throws JBIException {
        this.context = compContext;
        this.log = this.context.getLogger("", this.getResourceBundleName());

        this.log.fine("Component initialization in progress...");

        this.channel = this.context.getDeliveryChannel();
        this.componentConfiguration = this.createComponentConfiguration();
        this.runtimeConfiguration = this.createRuntimeConfiguration();
        this.asyncMessageHandler = new AsyncMessageHandler(this.log);

        // Necessary resources must be instantiated only once during the
        // life-cycle of the component
        if (this.suManager == null) {
            // TODO: interceptors and their JMX notifiers must be properly
            // handled against the life-cycle shutdown/init phases
            this.interceptors = this.loadInterceptors();
            this.interceptorManager = this.createInterceptorManager();

            this.suManager = this.createServiceUnitManager();
        }

        this.log.fine("Calling specific initialization...");
        this.doInit();
        this.log.fine("Specific initialization done.");

        // Since the worker manager instantiates the JBIListeners, we must
        // initialize it after the component specific initialization. This is
        // because the JBIListener implementation can access to the specific
        // component implementation which is done in the previous this.doInit()
        // call.
        this.workerManager = this.createWorkerManager();
        this.workerManager.init();
        this.acceptorManager = this.createJBIAcceptorManager();
        this.acceptorManager.init();

        this.log.log(Level.INFO, "Component initialized");
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#isExchangeWithConsumerOkay(javax.jbi.servicedesc.ServiceEndpoint,
     *      javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithConsumerOkay(final ServiceEndpoint endpoint,
            final MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#isExchangeWithProviderOkay(javax.jbi.servicedesc.ServiceEndpoint,
     *      javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithProviderOkay(final ServiceEndpoint endpoint,
            final MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#resolveEndpointReference(org.w3c.dom.DocumentFragment)
     */
    public ServiceEndpoint resolveEndpointReference(final DocumentFragment epr) {
        return null;
    }

    public void setServiceUnitListener(final ServiceUnitListener serviceUnitListener) {
        this.serviceUnitListener = serviceUnitListener;
    }

    public void setServiceUnitExtensionsValidator(
            final ServiceUnitExtensionsValidator serviceUnitExtensionsValidator) {
        this.serviceUnitExtensionsValidator = serviceUnitExtensionsValidator;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#shutDown()
     */
    public final void shutDown() throws JBIException {
        this.log.log(Level.INFO, "Shutdown Component");
        this.doShutdown();

        // remove unnecessary resources
        this.channel = null;
        this.acceptorManager = null;
        this.workerManager = null;
        this.asyncMessageHandler.shutdown();
        this.asyncMessageHandler = null;
        this.runtimeConfiguration = null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#start()
     */
    public final void start() throws JBIException {
        this.log.log(Level.INFO, "Start Component");
        this.registerPerformanceNotifierMBean();

        if (this.channel == null) {
            this.channel = this.context.getDeliveryChannel();
        }

        this.registerComponentInformationMBean();
        this.registerRuntimeConfigMBean();
        this.registerInterceptorManagerMBean();
        this.workerManager.start();
        this.acceptorManager.start();

        this.log.fine("Calling specific start...");
        this.doStart();
        this.log.fine("Specific start done.");
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#stop()
     */
    public final void stop() throws JBIException {
        this.log.log(Level.INFO, "Stop Component");
        // FIXME : call terminate or close?
        this.unregisterPerformanceNotifierMBean();
        this.unregisterComponentInformationMBean();
        this.unregisterRuntimeConfigurationMBean();
        this.unregisterInterceptorManagerMBean();
        // stop first the acceptor Manager as it aggregates the worker manager
        this.acceptorManager.stop();
        this.workerManager.stop();

        this.log.fine("Calling specific stop...");
        this.doStop();
        this.log.fine("Specific stop done.");
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doInit(). This method is called at the end of the component
     * classical initialization. When this method is called, the component's
     * delivery channel, context, logger and service unit manager are available.
     * 
     */
    protected void doInit() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doStart(). This method is called at the end of the component
     * classical start.
     * 
     */
    protected void doStart() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doStop(). This method is called at the end of the component
     * classical stop.
     * 
     */
    protected void doStop() throws JBIException {
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.doShutdown(). This method is called at the end of the
     * component classical shutdown.
     * 
     */
    protected void doShutdown() throws JBIException {
    }

    public ComponentConfiguration getComponentConfiguration() {
        return this.componentConfiguration;
    }

    /**
     * Load the interceptors
     * 
     * @param configuration
     */
    protected final Map<String, Interceptor> loadInterceptors() {
        Map<String, Interceptor> interceptors = null;
        final InterceptorBuilder builder = InterceptorBuilder.getInstance();
        try {
            final List<InterceptorConfiguration> list = builder.loadComponentInterceptors(this
                    .getContext().getInstallRoot());
            interceptors = this.instanciateInterceptors(list);
        } catch (final PEtALSCDKException e) {
            e.printStackTrace();
        }
        return interceptors;
    }

    /**
     * Load the accept interceptors
     * 
     * @return
     */
    protected final Map<String, Interceptor> instanciateInterceptors(
            final List<InterceptorConfiguration> interceptors) {
        final Map<String, Interceptor> loaded = new HashMap<String, Interceptor>();

        if (interceptors != null) {
            final InterceptorBuilder builder = InterceptorBuilder.getInstance();
            for (final InterceptorConfiguration interceptorConfig : interceptors) {
                try {
                    this.getLogger().info("Loading interceptor : " + interceptorConfig.getName());
                    final Interceptor itc = builder.loadInstance(interceptorConfig.getClazz());
                    itc.init(this, interceptorConfig.getName(), interceptorConfig.isActive(),
                            interceptorConfig.getParameters(), this.getLogger());
                    loaded.put(itc.getName(), itc);
                } catch (final PEtALSCDKException e) {
                    this.getLogger().log(
                            Level.WARNING,
                            "Could not load or initialize interceptor '"
                                    + interceptorConfig.getName() + "' from class '"
                                    + interceptorConfig.getClazz()
                                    + "', this interceptor is skipped");
                }
            }
        }
        return loaded;
    }

    /**
     * 
     * @return
     */
    public JBIAcceptorManager getListenerManager() {
        return this.acceptorManager;
    }

    /**
     * Register the component information MBean in the JMX server
     * 
     * @throws JBIException
     */
    private void registerComponentInformationMBean() throws JBIException {
        try {
            final ObjectName componentInformationName = this.context.getMBeanNames()
                    .createCustomComponentMBeanName(
                            "component_information_" + this.context.getComponentName());

            this.componentInformationMBean = new ComponentInformationMBean(componentConfiguration,
                    componentInformationName);
            final MBeanServer server = this.context.getMBeanServer();

            if (server == null) {
                throw new JBIException("MBean server is null !");
            }

            if (server.isRegistered(componentInformationName)) {
                server.unregisterMBean(componentInformationName);
            }

            server.registerMBean(this.componentInformationMBean, componentInformationName);

        } catch (final InstanceNotFoundException e) {
            throw new JBIException(e);
        } catch (final MBeanRegistrationException e) {
            throw new JBIException(e);
        } catch (final InstanceAlreadyExistsException e) {
            throw new JBIException(e);
        } catch (final NotCompliantMBeanException e) {
            throw new JBIException(e);
        }
    }

    /**
     * Register the performance notifier in the JMX server
     * 
     * @throws JBIException
     */
    private void registerPerformanceNotifierMBean() throws JBIException {
        try {
            final ObjectName jmxPerformanceNotifierName = this.context.getMBeanNames()
                    .createCustomComponentMBeanName(
                            "performance_notifier_" + this.context.getComponentName());

            this.jmxPerformanceNotifierMBean = new PerformanceNotifierMBean(
                    this.componentConfiguration, jmxPerformanceNotifierName);

            final MBeanServer server = this.context.getMBeanServer();

            if (server == null) {
                throw new JBIException("MBean server is null !");
            }

            if (server.isRegistered(jmxPerformanceNotifierName)) {
                server.unregisterMBean(jmxPerformanceNotifierName);
            }

            server.registerMBean(this.jmxPerformanceNotifierMBean, jmxPerformanceNotifierName);

        } catch (final InstanceNotFoundException e) {
            throw new JBIException(e);
        } catch (final MBeanRegistrationException e) {
            throw new JBIException(e);
        } catch (final InstanceAlreadyExistsException e) {
            throw new JBIException(e);
        } catch (final NotCompliantMBeanException e) {
            throw new JBIException(e);
        }
    }

    /**
     * Register the {@link RuntimeConfigurationMBean}
     * 
     * @throws JBIException
     */
    private void registerRuntimeConfigMBean() throws JBIException {
        MBeanHelper helper = MBeanHelper.getInstance();
        final ObjectName runtimeConfigurationMBeanName = helper.createCustomMBeanName(this.context
                .getMBeanNames(), "runtime_configuration_" + this.context.getComponentName());
        this.runtimeConfiguration.setMBeanName(runtimeConfigurationMBeanName);
        helper.registerMBean(this.context.getMBeanServer(), runtimeConfigurationMBeanName,
                this.runtimeConfiguration);
    }

    /**
     * 
     * @throws JBIException
     */
    private void registerInterceptorManagerMBean() throws JBIException {
        MBeanHelper helper = MBeanHelper.getInstance();
        final ObjectName mbeanName = helper.createCustomMBeanName(this.context.getMBeanNames(),
                "interceptors_" + this.context.getComponentName());
        this.interceptorManager.setMBeanName(mbeanName);
        helper.registerMBean(this.context.getMBeanServer(), mbeanName, this.interceptorManager);
    }

    /**
     * Creates the {@link RuntimeConfiguration}
     * 
     */
    private RuntimeConfiguration createRuntimeConfiguration() {
        return new RuntimeConfiguration(this.componentConfiguration);
    }

    /**
     * Creates the {@link InterceptorManager}
     * 
     * @return
     */
    private InterceptorManager createInterceptorManager() {
        return new InterceptorManager(this);
    }

    /**
     * Unregister the runtime configuration MBean
     * 
     * @throws JBIException
     */
    private void unregisterRuntimeConfigurationMBean() throws JBIException {
        MBeanHelper.getInstance().unregisterMBean(this.context.getMBeanServer(),
                this.runtimeConfiguration.getMBeanName());
    }

    /**
     * 
     * @throws JBIException
     */
    private void unregisterInterceptorManagerMBean() throws JBIException {
        MBeanHelper.getInstance().unregisterMBean(this.context.getMBeanServer(),
                this.interceptorManager.getMBeanName());
    }

    /**
     * Unregister the performance notifier
     * 
     * @throws JBIException
     */
    private void unregisterPerformanceNotifierMBean() throws JBIException {
        try {
            final MBeanServer server = this.context.getMBeanServer();
            server
                    .unregisterMBean(this.jmxPerformanceNotifierMBean
                            .getJmxPerformanceNotifierName());
        } catch (final InstanceNotFoundException e) {
            throw new JBIException(e);
        } catch (final MBeanRegistrationException e) {
            throw new JBIException(e);
        }
    }

    /**
     * 
     * @throws JBIException
     */
    private void unregisterComponentInformationMBean() throws JBIException {
        try {
            final MBeanServer server = this.context.getMBeanServer();
            server.unregisterMBean(this.componentInformationMBean.getMbeanName());
        } catch (final InstanceNotFoundException e) {
            throw new JBIException(e);
        } catch (final MBeanRegistrationException e) {
            throw new JBIException(e);
        }
    }

    /**
     * Get the performance notifier
     * 
     * @return
     */
    public PerformanceNotifierMBean getJmxPerformanceNotifierMBean() {
        return this.jmxPerformanceNotifierMBean;
    }

    /**
     * Create a performance notification which the type RECEIVING_FROM_OUTSIDE
     * 
     * @return The performance notification to send.
     */
    public PerformanceNotification createPerformanceNotificationReceivingFromOutside(
            final String userData, final Exchange msgExchange) {
        return this.createPerformanceNotification(PerformanceNotification.RECEIVING_FROM_OUTSIDE,
                userData, msgExchange);
    }

    /**
     * Create a performance notification which the type SENDING_TO_OUTSIDE
     * 
     * @return The performance notification to send.
     */
    public PerformanceNotification createPerformanceNotificationSendingToOutside(
            final String userData, final Exchange msgExchange) {
        return this.createPerformanceNotification(PerformanceNotification.SENDING_TO_OUTSIDE,
                userData, msgExchange);
    }

    /**
     * Create a performance notification which the type RECEIVING_FROM_NMR
     * 
     * @return The performance notification to send.
     */
    public PerformanceNotification createPerformanceNotificationReceivingFromNMR(
            final String userData, final Exchange msgExchange) {
        return this.createPerformanceNotification(PerformanceNotification.RECEIVING_FROM_NMR,
                userData, msgExchange);
    }

    /**
     * Create a performance notification which the type SENDING_TO_NMR
     * 
     * @return The performance notification to send.
     */
    public PerformanceNotification createPerformanceNotificationSendingToNMR(final String userData,
            final Exchange msgExchange) {
        return this.createPerformanceNotification(PerformanceNotification.SENDING_TO_NMR, userData,
                msgExchange);
    }

    /**
     * Create a new notification
     * 
     * @param type
     * @param userData
     * @param msgExchange
     * @return
     */
    private PerformanceNotification createPerformanceNotification(final String type,
            final String userData, final Exchange msgExchange) {
        PerformanceNotification performanceNotification = null;

        if (this.componentConfiguration.isPerformanceNotificationsEnabled()) {
            performanceNotification = this.jmxPerformanceNotifierMBean
                    .createPerformanceNotifications(System.currentTimeMillis(), this.context
                            .getComponentName(), type, userData, msgExchange);
        }
        return performanceNotification;
    }

    /**
     * Send a performance notification using the performance notifier
     * 
     * @param performanceNotification
     */
    public void sendPerformanceNotification(final PerformanceNotification performanceNotification) {
        if (this.componentConfiguration.isPerformanceNotificationsEnabled()) {
            this.jmxPerformanceNotifierMBean.send(performanceNotification);
        }
    }

    /**
     * Allows to extract extensions from a Provides or Consumes SU sub element.
     * If extensions contains place holders (references to external file
     * properties), they are automatically replaced by real value
     * 
     * @param extensibleElement
     *            a {@link ServiceUnitExtensibleElement} : {@link Consumes} or
     *            {@link Provides}
     */
    public Extensions getExtensions(ServiceUnitExtensibleElement extensibleElement) {
        Extensions extensions = new Extensions(extensibleElement.getExtensions());
        String propertiesFilePath = getComponentConfiguration().getPropertiesFile();
        try {
            extensions.setPetalsExtensions(PropertiesUtil.fillPlaceHolder(propertiesFilePath,
                    extensions.getPetalsExtensions()));
        } catch (PEtALSCDKException e) {
            getLogger().log(Level.WARNING, "Cannot fill place holder of SU extensions", e);
        }
        return extensions;
    }

    /**
     * 
     * @return
     */
    public AsyncMessageHandler getAsyncMessageHandler() {
        return asyncMessageHandler;
    }
}
