/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.InstallationContext;
import javax.management.ObjectName;

import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.util.ComponentPropertiesManager;
import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.component.framework.util.PropertiesUtil;
import org.ow2.petals.component.framework.util.StringHelper;
import static org.ow2.petals.component.framework.Constants.Component.PLACEHOLDER_PROPERTIES_FILE;

/**
 * A simple bootstrap implementation. This bootstrap just copy the JBI
 * extensions in a properties file that will be used by components.
 * 
 * @author ofabre chamerling - eBM WebSourcing
 * 
 */
public class Bootstrap implements javax.jbi.component.Bootstrap {

    private static final String LOGGER_BOOTSTRAP_PREFIX = "logger.bootstrap.";

    /**
     * The component installation context
     */
    protected InstallationContext installContext;

    /**
     * The component properties manager
     */
    protected ComponentPropertiesManager propertiesManager;

    private Logger logger;

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#cleanUp()
     */
    public void cleanUp() throws JBIException {
    }

    /**
     * Returns null by default as this class does not provides JMX extension
     * MBean
     * 
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#init(javax.jbi.component.InstallationContext)
     */
    public void init(InstallationContext installContext) throws JBIException {
        this.installContext = installContext;
        this.logger = Logger.getLogger(LOGGER_BOOTSTRAP_PREFIX + installContext.getComponentName());
        try {
            doInit(installContext);
        } catch (Exception e) {
            throw new JBIException("Initialization exception", e);
        }
    }

    /**
     * The component properties are loaded in this order :
     * <ol>
     * <li>Initialize all properties with their default values,</li>
     * <li>Override with values defined in the JBI descriptor,</li>
     * <li>Override with previously persisted values,</li>
     * <li>All placeholders of the JBI descriptor are overriden by their values
     * defined in the external file,</li>
     * <li>Add additional properties defined in the component.</li>
     * </ol>
     * 
     * @param installContext
     */
    private void doInit(InstallationContext installContext) {
        if (this.propertiesManager == null) {
            this.propertiesManager = new ComponentPropertiesManager();
            this.propertiesManager.setRootPath(installContext.getInstallRoot());

            // 1. get the default values
            this.propertiesManager.initDefaultValues();

            // 2. get the properties from the JBI extensions. Will override the
            // default ones
            org.ow2.petals.jbi.descriptor.Extensions descExtensions = new org.ow2.petals.jbi.descriptor.Extensions();
            descExtensions.setDocumentFragment(installContext.getInstallationDescriptorExtension());
            Extensions extensions = new Extensions(descExtensions);
            this.propertiesManager.getProperties().putAll(extensions.getPetalsExtensions());

            // put the parameter names in the properties for future use...
            Set<String> parameterNames = extensions.getParameters();
            Map<String, String> parameters = new HashMap<String, String>();
            Iterator<String> paramIterator = parameterNames.iterator();
            while (paramIterator.hasNext()) {
                String name = paramIterator.next();
                parameters.put(Constants.PARAMETER_PREFIX + name, name);
            }
            this.propertiesManager.getProperties().putAll(parameters);

            // 3. get the properties from an available properties file (from a
            // previous installation). Override the previous values.
            Properties persistedProperties = this.propertiesManager.retrievePersistedProperties();
            this.propertiesManager.getProperties().putAll(persistedProperties);

            // 4. Now try to replaced optional place holders (defined in the
            // property file) by real external property values. We use the
            // persisted name of the properties file is it exists
            String propertiesFilePath = persistedProperties.getProperty(PLACEHOLDER_PROPERTIES_FILE,
                    extensions.getValue(PLACEHOLDER_PROPERTIES_FILE));
            if (!StringHelper.isNullOrEmpty(propertiesFilePath)) {
                try {
                    Properties placeHolderProperties = PropertiesUtil.getFilledPlaceHolder(
                            propertiesFilePath, extensions.getPetalsExtensions());

                    // replace oldies in manager
                    this.propertiesManager.getProperties().putAll(placeHolderProperties);
                } catch (PEtALSCDKException e) {
                    logger
                            .log(Level.WARNING, "Cannot fill place holder of component extensions",
                                    e);
                }
            }

            // do additional configuration from component implementation. (this
            // method is overridden in the default bootstrap)
            doConfig();

            // last step, persist properties for component startup usage
            this.propertiesManager.save();
        }
    }

    /**
     * Do additional configuration during the bootstrap initialization phase.
     * This method is called at the end of the initialization and can be
     * overridden to provide specific additional configuration.
     * 
     */
    protected void doConfig() {
    }

    /**
     * 
     * @return
     */
    protected ComponentPropertiesManager getComponentPropertiesManager() {
        return this.propertiesManager;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#onInstall()
     */
    public void onInstall() throws JBIException {
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Bootstrap#onUninstall()
     */
    public void onUninstall() throws JBIException {
    }

    public InstallationContext getInstallContext() {
        return installContext;
    }

    public Logger getLogger() {
        return logger;
    }
}
