/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.HashMap;
import java.util.Map;

import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;

import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.jbi.descriptor.Provides;
import org.ow2.petals.jbi.descriptor.ServiceUnitExtensibleElement;

/**
 * This class is used when a JBI message is received on the Component from the
 * JBI environment. <br>
 * Service Engine case : <br>
 * The goal of this class is to process a service with the JBI message
 * information.<br>
 * For instance, a ServiceUnit has been deployed on the component, defining a
 * <code>provides</code> section.<br>
 * This section defines an Endpoint <em>"myService/jbiMyServiceEndpoint"</em>
 * and some property extensions.<br>
 * A <em>"myService/jbiMyServiceEndpoint"</em> has been activated on the JBI
 * environment.<br>
 * <br>
 * When a JBI consumer sends in the JBI environment a message to the
 * <em>"myService/jbiMyServiceEndpoint"</em> endpoint, this component received
 * the message, and calls the <code>onJBIMessage(exchange)</code> method.<br>
 * Then, the method has to process the message according to optional property
 * extensions (extensions of the provides that has activated the endpoint).
 *
 * Binding Component case : <br>
 * The goal of this class is to call an external service, related to this JBI
 * message. The address associated with the requested endpoint is handled by the
 * Extensions object of the provides that has activated the endpoint.<br>
 * For instance, a ServiceUnit has been deployed on the component, defining a
 * <code>provides</code> section.<br>
 * This section defines a relation between the external address
 * <em>"http://www.enterprise.com/webservices/myService"</em> (an address
 * attribute of the provides extensions), and an Endpoint
 * <em>"myService/jbiMyServiceEndpoint"</em>.<br>
 * A <em>"myService/jbiMyServiceEndpoint"</em> has been activated on the JBI
 * environment.<br>
 * <br>
 * When a JBI consumer sends in the JBI environment a message to the
 * <em>"myService/jbiMyServiceEndpoint"</em> endpoint, this component received
 * the message, and calls the <code>onJBIMessage(exchange)</code> method.<br>
 * Then, the method has to send the message to
 * <em>"http://www.enterprise.com/webservices/myService"</em>. <br/>
 *
 * @author ofabre - EBM Websourcing
 *
 */
public abstract class AbstractJBIListener extends AbstractListener {

    private Provides provides;

    private Map<ServiceUnitExtensibleElement, Extensions> extensionsMap;

    public AbstractJBIListener() {
        super();
        extensionsMap = new HashMap<ServiceUnitExtensibleElement, Extensions>();
    }

    /**
     * <p>
     * This method is used to process a JBI message received on the Component
     * from the JBI environment.
     * </p>
     *
     * <p>
     * This method can't throws Exceptions. You must managed Exception yourself
     * by using {@link Exchange} methods :<br/> -
     * setFault(Exception) : create a {@link Fault} with the given Exception and
     * set it in the {@link MessageExchange}<br/> - setError(Exception) : set
     * Error part of the {@link MessageExchange} and set {@link MessageExchange}
     * status to error. Exceptions handled by {@link MessageExchange}s must be
     * in the container classpath.
     * </p>
     *
     * <p>
     * Pattern's use cases :<br/> - InOnly : DONE status is automatically set at
     * the end of the message exchange process.<br/> - RobustInOnly : DONE
     * status is automatically set if no Fault found in exchange at the end of
     * the message exchange process.<br/> - InOut : DONE status is never set
     * (forbidden).<br/> - InOptOut : DONE status is automatically set if no
     * Fault or Out message found in exchange at the end of the message exchange
     * process.
     *
     * @param exchange
     *            the message exchange to process, wrapped to add some util
     *            methods. Not null
     *
     * @throws RuntimeException
     *             RuntimeException will be handle by the MessageExchange as
     *             Error part. So handled {@link RuntimeException}s must be in
     *             the classpath of your Petals node(s)
     */
    protected abstract void onJBIMessage(Exchange exchange);

    /*
     * (non-Javadoc)
     *
     * @see org.ow2.petals.component.framework.listener.AbstractListener#getExtensions()
     */
    public final Extensions getExtensions() {
        Extensions result = null;
        if (consumes != null) {
            if (!extensionsMap.containsKey(consumes)) {

                result = new Extensions(consumes.getExtensions());
                fillExtensionsPlaceHolders(result);

                extensionsMap.put(consumes, result);
            } else {
                result = extensionsMap.get(consumes);
            }
        } else if (provides != null) {
            if (!extensionsMap.containsKey(provides)) {
                result = new Extensions(provides.getExtensions());
                fillExtensionsPlaceHolders(result);
                extensionsMap.put(provides, result);
            } else {
                result = extensionsMap.get(provides);
            }
        }
        return result;
    }

    /**
     *
     * @return
     */
    public Provides getProvides() {
        return this.provides;
    }

    /**
     *
     * @param provides
     */
    public void setProvides(Provides provides) {
        this.provides = provides;
    }

}
