/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import static org.ow2.petals.component.framework.Constants.Component.MEP;
import static org.ow2.petals.component.framework.Constants.Component.OPERATION;
import static org.ow2.petals.component.framework.Constants.Component.TIMEOUT;

import java.net.URI;
import java.util.Enumeration;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.Constants;
import org.ow2.petals.component.framework.Constants.MEPConstants;
import org.ow2.petals.component.framework.Constants.Interceptor.Phase;
import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.mbean.PerformanceNotification;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.component.framework.util.ExchangeImpl;
import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.component.framework.util.PropertiesUtil;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.jbi.descriptor.Consumes;

/**
 * @author Christophe HAMERLING (chamerling) - eBM WebSourcing
 * @author ofabre - eBM WebSourcing
 * @author alouis - eBM WebSourcing
 */
public abstract class AbstractListener {

    /**
     * The JBI component itself
     */
    protected AbstractComponent component;

    /**
     * The consumed endpoint
     */
    protected Consumes consumes;

    /**
     * The message interceptor processor; Each listener have its own interceptor
     * processor in order to get better performances.
     */
    protected MessageInterceptorProcessor messageInterceptorProcessor;

    /**
     * The messageExchange factory used to create MessageExchanges
     */
    protected MessageExchangeFactory factory;

    /**
     * Creates a new instance of {@link AbstractListener}
     * 
     */
    public AbstractListener() {
        super();
    }

    /**
     * Initialize the listener.
     * 
     * @param component
     */
    public final void init(AbstractComponent component) {
        this.setComponent(component);
        this.messageInterceptorProcessor = new MessageInterceptorProcessor(this.component,
                getLogger());
        this.factory = getComponent().getChannel().createExchangeFactory();
    }

    /**
     * An util method creating a message exchange matching a pattern specified
     * in Extensions. Valid patterns are : "in-only", "robust-in-only",
     * "in-opt-out" or "in-out"
     * 
     * @param extensions
     *            a {@link Extensions}, must not be null
     * @return a message exchange matching the pattern given in extensions
     * @throws MessagingException
     *             if the message exchange can't be created, if the specified
     *             pattern isn't valid or if no pattern specified in extensions
     * @deprecated use createExchange(MEPConstants mep)
     */
    public Exchange createMessageExchange(Extensions extensions) throws MessagingException {
        if (extensions == null) {
            throw new NullPointerException("'extensions' parameter must be non null");
        }
        MEPConstants mep = retrievePattern(extensions);
        return createExchange(mep);
    }

    /**
     * Create an exchange implementation matching the given pattern.
     * 
     * @param mep
     *            a specific MEP pattern
     * @return an exchange implementation matching the given pattern
     * @throws MessagingException
     *             if the message exchange can't be created or if the specified
     *             pattern isn't valid
     */
    public Exchange createExchange(MEPConstants mep) throws MessagingException {
        if (mep == null) {
            throw new IllegalArgumentException("'mep' parameter must be non null");
        }
        MessageExchange exchange = factory.createExchange(mep.value());
        return new ExchangeImpl(exchange);
    }

    /**
     * Set operation to call if exists in extensions. If no extension can be
     * found or if operation value is "" or missing, operation isn't set.
     * 
     * @param exchange
     *            a {@link MessageExchange}, must not be null
     * @param extensions
     */
    protected void setOperationFromExtensions(final Exchange exchange, Extensions extensions) {
        if (exchange == null) {
            throw new NullPointerException("exchange must not be null");
        }
        if (extensions != null) {
            String operationName = extensions.getValue(OPERATION);
            if (!StringHelper.isNullOrEmpty(operationName)) {
                QName operation = QName.valueOf(operationName);
                exchange.setOperation(operation);
            }
        }
    }

    /**
     * Set message exchange properties from extensions. All extensions key
     * starting with "org.ow2.petals" are automatically added to message
     * exchange properties.
     * 
     * @param exchange
     *            a {@link Exchange}, must not be null
     * @param extensions
     */
    protected void setMessageExchangePropertiesFromExtensions(final Exchange exchange,
            final Extensions extensions) {
        if (exchange == null) {
            throw new NullPointerException("exchange must not be null");
        }
        if (extensions != null) {
            Properties petalsExtensions = extensions.getPetalsExtensions();
            for (Enumeration<Object> keys = petalsExtensions.keys(); keys.hasMoreElements();) {
                String name = (String) keys.nextElement();
                if (name.startsWith("org.ow2.petals")) {
                    exchange.setProperty(name, petalsExtensions.getProperty(name));
                }
            }
        }
    }

    /**
     * Retrieve the MEP from the extensions.
     * 
     * @param extensions
     * @return mep specified in extensions
     * @throws MessagingException
     *             MEP pattern specified in consumes extension isn't valid or if
     *             no MEP pattern specified in extensions.
     */
    protected final MEPConstants retrievePattern(Extensions extensions) throws MessagingException {
        MEPConstants mep = null;
        String patternShortcut = extensions.getPetalsExtensions().getProperty(MEP);
        if (!StringHelper.isNullOrEmpty(patternShortcut)) {
            if (Constants.MEPShortcuts.ROBUST_IN_ONLY.equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
            } else if (Constants.MEPShortcuts.IN_OPTIONAL_OUT.equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
            } else if (Constants.MEPShortcuts.IN_OUT.equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.IN_OUT_PATTERN;
            } else if (Constants.MEPShortcuts.IN_ONLY.equalsIgnoreCase(patternShortcut)) {
                mep = MEPConstants.IN_ONLY_PATTERN;
            } else {
                throw new MessagingException("MEP pattern specified in " + "consumes extension ("
                        + patternShortcut + ") isn't valid. " + "Valid patterns are : in-only, "
                        + "in-opt-out, robust-in-only, in-out.");
            }
        } else {
            throw new MessagingException("No MEP pattern specified in extensions");
        }
        return mep;
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.init(). This method is called after setting of all Listener
     * attributes
     * 
     */
    protected void init() {
    }

    /**
     * Call accept(MessageExchange exchange, long timeout) with a timeout of "0"
     * that means no timeout
     * 
     * @param exchange
     *            the exchange that has been sent in an asynchronous way
     * @return the response {@link MessageExchange} to an asynchronous service
     *         call
     * @throws InterruptedException
     * @throws PEtALSCDKException
     *             if this method is called despite the fact that response
     *             reception has been acknowledged
     */
    public Exchange accept(Exchange exchange) throws InterruptedException, PEtALSCDKException {
        return this.accept(exchange, 0);
    }

    /**
     * This method blocks until response to the given {@link MessageExchange}
     * reaches the component. A timeout must be specified. Waiting stops when
     * timeout occured.
     * 
     * @param exchange
     *            the exchange that has been sent in an asynchronous way
     * @param timeout
     *            waiting message return until timeout occured. Timeout in
     *            milliseconds. "0" means no timeout
     * @return the response {@link MessageExchange} to an asynchronous service
     *         call. Can be null if timeout occured.
     * @throws InterruptedException
     * @throws PEtALSCDKException
     *             if this method is called despite the fact that response
     *             reception has been acknowledged
     */
    public Exchange accept(Exchange exchange, long timeout) throws InterruptedException,
            PEtALSCDKException {
        // Try to find the response in the synchronizer (response arrives before
        // the call to accept)
        component.getLogger().finest(
                "Wait for a response to exchange of id: " + exchange.getExchangeId());
        MessageExchangeSynchronizer exchangeSynchronizer = this.component.getAsyncMessageHandler()
                .getAsynchronousMessage(exchange.getExchangeId());
        Exchange responseExchange = null;
        if (exchangeSynchronizer != null) {
            responseExchange = exchangeSynchronizer.getExchange();
            if (responseExchange == null) {
                // No response arrived, so block until response arrival
                responseExchange = exchangeSynchronizer.accept(timeout);
            }
            exchangeSynchronizer.setAccepted(true);
        } else {
            throw new PEtALSCDKException("You are trying to accept a response for"
                    + " an exchange not synchronized");
        }
        component.getLogger().finest(
                "Asynchronous response arrived for exchange with id : " + exchange.getExchangeId());
        return responseExchange;
    }

    /**
     * This method must be called to enable the CDK to automatically send an
     * acknowledgment to the provider of the service, at the end of a CDK
     * managed asynchronous exchange. It automatically set exchange status to
     * DONE if exchange pattern is RobustInOnly or InOut or if exchange pattern
     * is InOptOut and no fault is set.
     * 
     * @param exchange
     *            the message exchange originating the response synchronization
     * @throws MessagingException
     *             if an error occurs during acknowledgment
     * @throws PEtALSCDKException
     *             if you try close exchange before accepting response or for an
     *             unsynchronized exchange
     */
    public void closeExchange(Exchange exchange) throws MessagingException, PEtALSCDKException {
        MessageExchangeSynchronizer exchangeSynchronizer = this.component.getAsyncMessageHandler()
                .getAsynchronousMessage(exchange.getExchangeId());
        if (exchangeSynchronizer != null) {
            if (exchangeSynchronizer.isAccepted() && !exchangeSynchronizer.isTimeout()) {
                Exchange returnedMessage = exchangeSynchronizer.getExchange();
                processAck(returnedMessage);
                exchangeSynchronizer.ackResponseProcessed();
                component.getChannel().send(((ExchangeImpl) returnedMessage).getMessageExchange());
                component.getLogger()
                        .finest("Close exchange with id : " + exchange.getExchangeId());
            } else {
                throw new PEtALSCDKException("You are trying to close an exchange despite "
                        + "the fact that you have not received response "
                        + "(call accept() method before " + "or specify a longest timeout)");
            }
        } else {
            throw new PEtALSCDKException("You are trying to close an exchange not synchronized");
        }
    }

    private void processAck(Exchange exchange) throws MessagingException {
        if (exchange.isRobustInOnlyPattern() || exchange.isInOutPattern()) {
            if (exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
        } else if (exchange.isInOptionalOutPattern()) {
            if (exchange.getFault() == null && exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
        }
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/> Process MessageExchange interceptors before
     * sending message. <br/> No synchronization mecanism activated
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public void send(final Exchange exchange) throws MessagingException {

        if (exchange == null) {
            throw new NullPointerException("You must provide a non null MessageExchange");
        }

        interceptExchange(exchange);

        component.getLogger().finest(
                "Send an asynchronous exchange with id : " + exchange.getExchangeId());
        getDeliveryChannel().send(((ExchangeImpl) exchange).getMessageExchange());
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/> Process MessageExchange interceptors before
     * sending message. <br/> Add a synchronization mecanism that enable usage
     * of a blocking accept(messageId) method that will be unlock when the
     * response message with the same id reaches the component
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public void sendAsync(final Exchange exchange) throws MessagingException {

        if (exchange == null) {
            throw new NullPointerException("You must provide a non null MessageExchange");
        }

        // Register a synchroniser
        MessageExchangeSynchronizer exchangeSynchronizer = new MessageExchangeSynchronizer();
        this.component.getAsyncMessageHandler().addAsynchronousMessage(exchange.getExchangeId(),
                exchangeSynchronizer);

        this.send(exchange);
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * synchronous way.<br/> Process MessageExchange interceptors before
     * sending message. <br/> Add performance notification
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @param timeoutMS
     *            timeout period, in milliseconds (0 means no timeout); must be
     *            greater than or equal to zero
     * @return true if the message has been sent (if a timeout occurs, return
     *         false)
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange, long timeoutMS) throws MessagingException {

        if (exchange == null) {
            throw new NullPointerException("You must provide a non null MessageExchange");
        }

        this.interceptExchange(exchange);

        // Notification creation
        PerformanceNotification performanceNotification = this.component
                .createPerformanceNotificationSendingToNMR(this
                        .getPerformanceNotificationUserData(exchange), exchange);

        component.getLogger().finest(
                "Send a synchronous exchange with id : " + exchange.getExchangeId());
        boolean bRet = getDeliveryChannel().sendSync(
                ((ExchangeImpl) exchange).getMessageExchange(), timeoutMS);

        // Notification sending
        this.component.sendPerformanceNotification(performanceNotification);

        this.interceptExchangeResponse(exchange);

        return bRet;
    }

    /**
     * This method should be overriden if you wish to set custom user data to
     * your performance notification. By default, the user data is set to the
     * mesasge exchange endpoint name if exists, else with the service name,
     * else with the interface name.
     * 
     * @param exchange
     * @return
     */
    protected String getPerformanceNotificationUserData(final Exchange exchange) {
        String sRet = null;
        if (exchange.getEndpoint() != null) {
            sRet = exchange.getEndpoint().toString();
        }

        if (sRet == null) {
            if (exchange.getService() != null) {
                sRet = exchange.getService().toString();
            }
        }

        if (sRet == null) {
            if (exchange.getInterfaceName() != null) {
                sRet = exchange.getInterfaceName().toString();
            }
        }

        return sRet;
    }

    /**
     * Call sendSync(MessageExchage exchange, long timeoutMS) with a timeout
     * value found in provides or consumes extensions or "0" if not found
     * 
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange) throws MessagingException {
        long timeoutMS = 0;
        Extensions extensions = getExtensions();
        if (extensions != null) {
            // retrieve timeout
            String timeout = extensions.getPetalsExtensions().getProperty(TIMEOUT, "0");
            if (StringHelper.isNullOrEmpty(timeout)) {
                timeoutMS = Long.parseLong("0");
            } else {
                timeoutMS = Long.parseLong(timeout);
            }
        }
        return sendSync(exchange, timeoutMS);
    }

    /**
     * Call the Interceptors before sending the exchange
     * 
     * @param exchange
     */
    protected void interceptExchange(final Exchange exchange) {
        if (consumes != null) {
            this.messageInterceptorProcessor.intercept(exchange, consumes, Phase.SEND);
        }
    }

    /**
     * Call the Interceptors on exchange response
     * 
     * @param exchange
     */
    protected void interceptExchangeResponse(final Exchange exchange) {
        if (consumes != null) {
            this.messageInterceptorProcessor.intercept(exchange, consumes, Phase.SEND_RESPONSE);
        }
    }

    /**
     * Gets the extensions. Find them in consumes or provides nodes
     * 
     * @return Extensions, can be null if no extensions found
     */
    public abstract Extensions getExtensions();

    /**
     * 
     * @return
     */
    @Deprecated
    public ComponentContext getComponentContext() {
        return component.getContext();
    }

    /**
     * 
     * @return
     */
    @Deprecated
    public DeliveryChannel getDeliveryChannel() {
        return component.getChannel();
    }

    /**
     * 
     * @return
     */
    public Logger getLogger() {
        return component.getLogger();
    }

    /**
     * 
     * @param component
     */
    private void setComponent(AbstractComponent component) {
        this.component = component;
    }

    /**
     * 
     * @return
     */
    public AbstractComponent getComponent() {
        return this.component;
    }

    /**
     * @return the consumes
     */
    public Consumes getConsumes() {
        return this.consumes;
    }

    /**
     * @param consumes
     *            the consumes to set
     */
    public void setConsumes(Consumes consumes) {
        this.consumes = consumes;
    }

    /**
     * @return the messageInterceptorProcessor
     */
    public MessageInterceptorProcessor getMessageInterceptorProcessor() {
        return this.messageInterceptorProcessor;
    }

    protected void fillExtensionsPlaceHolders(Extensions extensions) {
        // Now try to replaced optionnal place holders by real external
        // property values
        String propertiesFilePath = component.getComponentConfiguration().getPropertiesFile();
        if (propertiesFilePath != null && !"".equals(propertiesFilePath)) {
            try {
                extensions.setPetalsExtensions(PropertiesUtil.fillPlaceHolder(propertiesFilePath,
                        extensions.getPetalsExtensions()));
            } catch (PEtALSCDKException e) {
                getLogger().log(Level.WARNING, "Cannot fill place holder of SU extensions", e);
            }
        }
    }

    /**
     * Create a message exchange with all the information set in the specified
     * CONSUME <br>- address (interface, service, endpoint) <br>- MEP <br>-
     * operation <br>- PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *             if the message exchange can't be created or if no pattern
     *             specified in extensions
     */
    public Exchange createConsumeExchange(Consumes consumes) throws MessagingException {
        Extensions extensions = new Extensions(consumes.getExtensions());
        fillExtensionsPlaceHolders(extensions);
        MEPConstants mep = retrievePattern(extensions);
        Exchange exchange = createConsumeExchange(consumes, mep.value());
        return exchange;
    }

    /**
     * Create a message exchange with all the information set in the specified
     * CONSUME, with the given exchange pattern <br>- address (interface,
     * service, endpoint) <br>- operation <br>- PEtALS exchange' meta-data
     * (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *             if the message exchange can't be created or if no pattern
     *             specified in extensions
     */
    public Exchange createConsumeExchange(Consumes consumes, URI mep) throws MessagingException {
        Extensions extensions = new Extensions(consumes.getExtensions());
        fillExtensionsPlaceHolders(extensions);
        Exchange exchange = new ExchangeImpl(factory.createExchange(mep));
        setOperationFromExtensions(exchange, extensions);
        setMessageExchangePropertiesFromExtensions(exchange, extensions);
        setExchangeDestinationForConsumes(exchange, consumes);
        return exchange;
    }

    /**
     * Fill the exchange with the service information (itf,src, and ep) from the
     * consume extension
     * 
     * @param exchange
     */
    private void setExchangeDestinationForConsumes(final Exchange exchange, final Consumes consumes) {
        // Retrieve possible target endpoint
        ServiceEndpoint ep = null;
        if (consumes.getEndpointName() != null && consumes.getServiceName() != null) {
            ep = getComponent().getContext().getEndpoint(consumes.getServiceName(),
                    consumes.getEndpointName());
        }
        // set exchange information
        exchange.setInterfaceName(consumes.getInterfaceName());
        exchange.setService(consumes.getServiceName());
        exchange.setEndpoint(ep);
    }

}
