/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;
import java.util.logging.Logger;

/**
 * An asynchronous message handler.
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 3.0
 * 
 */
public class AsyncMessageHandler {

    private Logger logger;

    private Timer timer;

    private Map<String, MessageExchangeSynchronizer> asynchronousMessages;

    /**
     * 
     * @param logger
     */
    public AsyncMessageHandler(final Logger logger) {
        this.logger = logger;
        this.asynchronousMessages = Collections
                .synchronizedMap(new HashMap<String, MessageExchangeSynchronizer>());
        cleanMESynchronizers();
    }

    /**
     * Add an asynchronous message to the asynchronous messages map.
     * 
     * @param id
     * @param exchangeSynchronizer
     */
    public void addAsynchronousMessage(String id, MessageExchangeSynchronizer exchangeSynchronizer) {
        synchronized (this.asynchronousMessages) {
            logger.finest("Add synchronizer for exchange id: " + id);
            this.asynchronousMessages.put(id, exchangeSynchronizer);
        }
    }

    /**
     * Get an asynchronous message from its id.
     * 
     * @param id
     * @return null if the asynchronous message is not found in the map
     */
    public MessageExchangeSynchronizer getAsynchronousMessage(String id) {
        return asynchronousMessages.get(id);
    }

    /**
     * Remove an asynchronous message from the map
     * 
     * @param id
     */
    public void removeAsynchronousMessage(String id) {
        synchronized (this.asynchronousMessages) {
            logger.finest("Remove synchronizer for exchange id: " + id);
            this.asynchronousMessages.remove(id);
        }
    }

    /**
     * Shutdown the AsyncMessageHandler. Remove its internal timer
     */
    public void shutdown() {
        timer.cancel();
    }

    /**
     * Periodically clean ME synchronizer map. A synchronizer can be removed
     * only if ME response has been accepted and provider has been acknowledged
     */
    private void cleanMESynchronizers() {
        timer = new Timer();
        MESynchronizerCleaner cleaner = new MESynchronizerCleaner();
        timer.schedule(cleaner, 0, 2000);
    }

    /**
     * The clearer task
     */
    public class MESynchronizerCleaner extends TimerTask {
        @Override
        public void run() {
            ArrayList<MessageExchangeSynchronizer> synchronizers = new ArrayList<MessageExchangeSynchronizer>(
                    asynchronousMessages.values());
            for (MessageExchangeSynchronizer synchronizer : synchronizers) {
                if (synchronizer.isAcknowledged()) {
                    removeAsynchronousMessage(synchronizer.getExchange().getExchangeId());
                }
            }
        }
    }
}
