/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.management.AttributeChangeNotification;
import javax.management.Notification;
import javax.management.NotificationListener;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.ComponentConfiguration;
import org.ow2.petals.component.framework.mbean.RuntimeConfiguration;
import org.ow2.petals.component.framework.mbean.RuntimeConfigurationMBean;

/**
 * This class helps a Component to listen to {@link MessageExchange}s on the
 * DeliveryChannel (accept() method). It acts as a ListenerManager. It starts or
 * stops a pool of thread that listen to new incoming JBI
 * {@link MessageExchange}s<p/>
 * <p>
 * 
 * @version $Revision: $ $Date: $
 * @since 3.0
 * @author alouis, ofabre
 */
public class JBIAcceptorManager implements PoolManager {

    private Logger logger;

    private ThreadPoolExecutor acceptorPool;

    private WorkerPoolManager workerManager;

    private ComponentConfiguration configuration;

    private AbstractComponent component;

    private RuntimeConfiguration runtimeConfiguration;

    /**
     * 
     */
    private NotificationListener listener = new NotificationListener() {
        /*
         * (non-Javadoc)
         * 
         * @see javax.management.NotificationListener#handleNotification(javax.management.Notification,
         *      java.lang.Object)
         */
        public void handleNotification(Notification notification, Object obj) {
            if (notification instanceof AttributeChangeNotification) {
                AttributeChangeNotification attrNotif = (AttributeChangeNotification) notification;
                String attrName = attrNotif.getAttributeName();
                logger.log(Level.FINE, "Notification received for attribute : " + attrName);
                if (attrName.equals(RuntimeConfiguration.ACCEPTOR_THREADS)) {
                    Integer newVal = (Integer) (attrNotif.getNewValue());
                    setCorePoolSize(newVal.intValue());
                }
            }
        }
    };

    /**
     * Creates a new instance of {@link JBIAcceptorManager}
     * 
     * @param channel
     * @param configuration
     *            The configuration of component
     * @param componentName
     *            Name of the component. Used to define the Thread name.
     * @param log
     * @param suManager
     * @param listener
     */
    public JBIAcceptorManager(AbstractComponent component,
            final ComponentConfiguration configuration, final WorkerPoolManager workerManager,
            final RuntimeConfiguration runtimeConfiguration, Logger log) {
        super();
        this.component = component;
        this.configuration = configuration;
        this.workerManager = workerManager;
        this.runtimeConfiguration = runtimeConfiguration;

        this.runtimeConfiguration.addNotificationListener(listener, null, null);

        if (log != null) {
            this.logger = log;
        } else {
            this.logger = LogManager.getLogManager().getLogger(
                    component.getContext().getComponentName());
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.listener.PoolManager#init()
     */
    public void init() throws JBIException {
        createAcceptorPool();
    }

    /**
     * Start threads which listen to {@link MessageExchange}s
     * 
     * @throws JBIException
     * 
     */
    public void start() throws JBIException {
        this.startAcceptorPool();
    }

    /**
     * Create the JBI acceptors pool.
     */
    protected void createAcceptorPool() {
        // Note : since core-pool-size and maximum-pool-size are the same, the
        // pool size is fixed.
        this.acceptorPool = new ThreadPoolExecutor(configuration.getAcceptorSize(), configuration
                .getAcceptorSize(), 60, TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>(1024),
                new JBIAcceptorThreadFactory(this.component, this.workerManager, this.logger));
    }

    /**
     * 
     */
    protected void startAcceptorPool() {
        this.acceptorPool.prestartAllCoreThreads();
    }

    /**
     * Stop threads that listen to {@link MessageExchange}s
     * 
     */
    public void stop() {
        // the executed task are hanging on the componentContext.accept()
        // method, so force the stop of these tasks directly by invoking
        // shutdownNow() instead of shutdown()
        this.acceptorPool.shutdownNow();
    }

    /**
     * Called on JMX notification
     * 
     * @param threadCount
     */
    private void setCorePoolSize(int threadCount) {
        if (acceptorPool != null) {
            acceptorPool.setCorePoolSize(threadCount);
        }
    }

    /**
     * The Acceptor factory used to create new Acceptors
     * 
     * @author Christophe HAMERLING - eBM WebSourcing
     * @since 3.0
     * 
     */
    private class JBIAcceptorThreadFactory implements ThreadFactory {
        private Logger logger;

        private AbstractComponent component;

        private RuntimeConfigurationMBean runtimeConfiguration;

        private ThreadPoolExecutor workersPool;

        private GenericObjectPool jbiListenerPool;

        final AtomicInteger threadNumber = new AtomicInteger(1);

        /**
         * 
         * @param component
         * @param workersPool
         * @param jbiListenerPool
         * @param logger
         */
        public JBIAcceptorThreadFactory(final AbstractComponent component,
                final WorkerPoolManager workerPoolManager, final Logger logger) {
            this.component = component;
            this.workersPool = workerPoolManager.getThreadPool();
            this.jbiListenerPool = workerPoolManager.getWorkersPool();
            this.logger = logger;
        }

        /*
         * (non-Javadoc)
         * 
         * @see java.util.concurrent.ThreadFactory#newThread(java.lang.Runnable)
         */
        public Thread newThread(Runnable r) {
            String name = "JBI Acceptor Thread #"
                    + Integer.toString(threadNumber.getAndIncrement());

            this.logger.log(Level.FINE, "Creating a new thread for JBIAcceptor Factory : " + name);

            MessageExchangeAcceptor meaProcessor = new MessageExchangeAcceptor(this.component,
                    this.jbiListenerPool);
            meaProcessor.setName(name);
            meaProcessor.setLogger(logger);
            meaProcessor.setWorkersPool(workersPool);

            return meaProcessor;
        }
    }
}
