/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.Constants.IgnoredStatusConstants;
import org.ow2.petals.component.framework.Constants.Interceptor.Phase;
import org.ow2.petals.component.framework.mbean.PerformanceNotification;
import org.ow2.petals.component.framework.su.ServiceUnitManager;
import org.ow2.petals.component.framework.util.Exchange;
import org.ow2.petals.component.framework.util.ExchangeImpl;
import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.jbi.descriptor.Consumes;
import org.ow2.petals.jbi.descriptor.Provides;

/**
 * Processor for received MessageExchanges. This processor can be overridden to
 * define a specific strategy to accept messages from the DeliveryChannel. See
 * <code>acceptMessageExchange()</code>. <br>
 * Override the <code>createMessageExchangeProcessor</code> method in the
 * <code>Component</code> class to set your processor.
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.3
 * @author alouis, ofabre - EBM Websourcing
 * @author Christophe HAMERLING - eBM WebSourcing
 */
public class MessageExchangeProcessor extends Thread {

    /**
     * The component's logger
     */
    protected Logger logger;

    /**
     * The component's delivery channel
     */
    private DeliveryChannel channel;

    private AbstractComponent component;

    private AbstractJBIListener listener;

    private GenericObjectPool jbiPool;

    /**
     * The component's service unit manager
     */
    private ServiceUnitManager suManager;

    private ExchangeImpl exchange;

    /**
     * Creates a processor that listen to new {@link MessageExchange} on the
     * specified channel. Exchanges are ignored depending on the specified
     * exchange-status to ignore.
     * 
     * @param component
     * @param jbiPool
     * @param exchange
     */
    public MessageExchangeProcessor(AbstractComponent component, GenericObjectPool jbiPool,
            ExchangeImpl exchange) {
        this.component = component;
        this.jbiPool = jbiPool;
        this.channel = component.getChannel();
        this.suManager = (ServiceUnitManager) component.getServiceUnitManager();
        this.exchange = exchange;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public void run() {
        if (exchange != null) {
            if (!this.isIgnored(exchange)) {
                PerformanceNotification performanceNotification = null;
                try {
                    this.listener = (AbstractJBIListener) jbiPool.borrowObject();
                    performanceNotification = this.component
                            .createPerformanceNotificationReceivingFromNMR(this
                                    .getPerformanceNotificationUserData(), exchange);

                    this.process(exchange);

                } catch (Throwable e) {
                    this.logger.log(Level.SEVERE, e.getMessage(), e);
                    exchange.setError(new Exception(e));
                    try {
                        this.channel.send(exchange.getMessageExchange());
                    } catch (MessagingException e1) {
                        this.logger.log(Level.SEVERE, "Can't send Provider Error "
                                + "MessageExchange to consumer", e);
                    }
                } finally {
                    this.component.sendPerformanceNotification(performanceNotification);
                    try {
                        jbiPool.returnObject(this.listener);
                    } catch (Exception e) {
                        logger.warning("Can not send back listener to pool...");
                    }
                }
            }
        }
    }

    public void setJBIListener(AbstractJBIListener listener) {
        this.listener = listener;
    }

    public void setLogger(Logger logger) {
        this.logger = logger;
    }

    /**
     * This method should be overriden if you wish to set custom user data to
     * your performance notification.
     * 
     * @return Perfomance notification user data.
     */
    // TODO: Use the caller name as user data.
    protected String getPerformanceNotificationUserData() {
        return "";
    }

    /**
     * depending on the IgnoredStatus, tells if the exchange has to be ignored.
     * If the exchange is null, return true.
     * 
     * @param exchange
     * @return
     */
    protected boolean isIgnored(final Exchange exchange) {
        boolean result = false;

        if (((ExchangeImpl) exchange).getMessageExchange() == null) {
            result = true;
        } else {
            IgnoredStatusConstants status = this.component.getComponentConfiguration()
                    .getIgnoredStatus();
            if (exchange.isDoneStatus()) {
                result = IgnoredStatusConstants.DONE_IGNORED == status
                        || IgnoredStatusConstants.DONE_AND_ERROR_IGNORED == status;
            } else if (exchange.isErrorStatus()) {
                result = IgnoredStatusConstants.ERROR_IGNORED == status
                        || IgnoredStatusConstants.DONE_AND_ERROR_IGNORED == status;
            }
        }
        return result;
    }

    /**
     * If the received exchange is a fault response from the consumer,
     * acknowledge it : status DONE
     * 
     * @param exchange
     *            The processed {@link MessageExchange}
     * @return true if the ack. of the Fault is done; false if no fault is
     *         present in the exchange
     * @throws MessagingException
     */
    private boolean ackFaultReception(final Exchange exchange) throws MessagingException {
        boolean result = false;

        if (exchange.getFault() != null) {
            this.listener.onJBIMessage(exchange);
            exchange.setDoneStatus();
            result = true;
        }
        return result;
    }

    /**
     * Get the consumes element for the given exchange
     * 
     * @param exchange
     * @return
     */
    private Consumes getConsumesFromExchange(Exchange exchange) {
        ServiceEndpoint endpoint = exchange.getEndpoint();
        return this.suManager.getConsumesFromDestination(endpoint.getEndpointName(), exchange
                .getService(), exchange.getInterfaceName());
    }

    /**
     * 
     * @param exchange
     * @return
     */
    private Provides getProvidesFromExchangeEndpoint(final Exchange exchange) {
        ServiceEndpoint ep = exchange.getEndpoint();
        Provides provides = null;
        if (ep != null) {
            provides = this.suManager.getProvidesFromEndpoint(ep);
        }
        return provides;
    }

    /**
     * Launch the component processing of the given message exchange
     * 
     * @param exchange
     *            the message exchange to process. Must be non null.
     * @throws InterruptedException
     * @throws JBIException
     *             if an error occured during the message exchange processing
     * @throws MessagingException
     *             impossible to process the exchange.
     */
    private void process(final Exchange exchange) throws MessagingException, InterruptedException {
        boolean exchangeTerminated = false;

        if (!exchange.isActiveStatus()) {
            exchangeTerminated = true;
        }

        if (exchange.isProviderRole()) {
            this.logger.finest("Process an exchange as PROVIDER with id : "
                    + exchange.getExchangeId());
            Provides provides = this.getProvidesFromExchangeEndpoint(exchange);

            if (provides == null) {
                throw new MessagingException("This endpoint (" + exchange.getEndpoint()
                        + ") is unknown on this component : " + this.component);
            }

            this.listener.setProvides(provides);

            try {
                this.listener.getMessageInterceptorProcessor().intercept(exchange, provides,
                        Phase.ACCEPT);

                this.processAsProvider(exchange);

                this.listener.getMessageInterceptorProcessor().intercept(exchange, provides,
                        Phase.ACCEPT_RESPONSE);

                if (!exchangeTerminated) {
                    //set the PEtALS dedicated properties to the exchange 
                    Properties properties = this.listener.getExtensions().getPetalsExtensions();
                    for (Object key : properties.keySet()) {
                        if (key.toString().startsWith("org.ow2.petals.")) {
                            exchange.setProperty(key.toString(), properties.get(key));
                        }
                    }
                    this.channel.send(((ExchangeImpl) exchange).getMessageExchange());
                }
            } finally {
                this.listener.setProvides(null);
            }

        } else {
            this.logger.finest("Process an exchange as CONSUMER with id : "
                    + exchange.getExchangeId());

            Consumes consumes = this.getConsumesFromExchange(exchange);
            // Notify a synchronizer if exist
            MessageExchangeSynchronizer exchangeSynchronizer = this.component
                    .getAsyncMessageHandler().getAsynchronousMessage(exchange.getExchangeId());

            this.listener.getMessageInterceptorProcessor().intercept(exchange, consumes,
                    Phase.SEND_RESPONSE);

            if (exchangeSynchronizer != null) {
                if (!exchangeSynchronizer.isTimeout()) {
                    exchangeSynchronizer.setExchange(exchange);
                    exchangeSynchronizer.notifyExchange();
                } else if (!exchangeTerminated) {
                    exchange.setError(new Exception("The response to an asynchronous exchange "
                            + "arrived after a timeout occured during accept"));
                    this.channel.send(((ExchangeImpl) exchange).getMessageExchange());
                }
            } else {
                this.listener.setConsumes(consumes);
                this.processAsConsumer(exchange);
                this.listener.setConsumes(null);
                if (!exchangeTerminated) {
                    this.channel.send(((ExchangeImpl) exchange).getMessageExchange());
                }
            }
        }
    }

    /**
     * Process the exchange as service consumer
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processAsConsumer(final Exchange exchange) throws MessagingException {
        if (exchange.isRobustInOnlyPattern()) {
            this.processRobustInOnlyAsConsumer(exchange);
        } else if (exchange.isInOnlyPattern()) {
            this.processInOnlyAsConsumer(exchange);
        } else if (exchange.isInOutPattern()) {
            this.processInOutAsConsumer(exchange);
        } else if (exchange.isInOptionalOutPattern()) {
            this.processInOptionalOutAsConsumer(exchange);
        } else {
            throw new MessagingException("MessageExchangePattern not recognized :"
                    + exchange.getPattern());
        }
    }

    /**
     * Process the exchange as service provider
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processAsProvider(final Exchange exchange) throws MessagingException {
        if (exchange.isInOnlyPattern()) {
            this.processInOnlyAsProvider(exchange);
        } else if (exchange.isRobustInOnlyPattern()) {
            this.processRobustInOnlyAsProvider(exchange);
        } else if (exchange.isInOutPattern()) {
            this.processInOutAsProvider(exchange);
        } else if (exchange.isInOptionalOutPattern()) {
            this.processInOptionalOutAsProvider(exchange);
        } else {
            throw new MessagingException("MessageExchangePattern not recognized :"
                    + exchange.getPattern());
        }
    }

    /**
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processInOnlyAsConsumer(Exchange exchange) throws MessagingException {
        if (exchange.isDoneStatus() || exchange.isErrorStatus()) {
            this.listener.onJBIMessage(exchange);
        } else {
            exchange.setError(new JBIException("Bad exchange state : must be Done or Error"));
        }
    }

    /**
     * Process "in-only" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws MessagingException
     *             When errors occur during processing
     */
    private void processInOnlyAsProvider(final Exchange exchange) throws MessagingException {
        this.listener.onJBIMessage(exchange);
        if (exchange.isActiveStatus()) {
            exchange.setDoneStatus();
        }
    }

    /**
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processInOptionalOutAsConsumer(Exchange exchange) throws MessagingException {
        this.listener.onJBIMessage(exchange);

        // TODO Delete the out message if the it still handled by the message
        // exchange

        // Set Status DONE if no Fault and ACTIVE Status
        if (exchange.getFault() == null && exchange.isActiveStatus()) {
            exchange.setDoneStatus();
        }
    }

    /**
     * Process "in-optional-out" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws MessagingException
     *             When errors occur during processing
     */
    private void processInOptionalOutAsProvider(Exchange exchange) throws MessagingException {
        if (!this.ackFaultReception(exchange)) {
            this.listener.onJBIMessage(exchange);

            // Set Status DONE if no Fault or Out messages and ACTIVE Status
            if (exchange.getFault() == null && (!(((ExchangeImpl) exchange).isOutMessage()))
                    && exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
        }
    }

    /**
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processInOutAsConsumer(Exchange exchange) throws MessagingException {
        try {
            this.listener.onJBIMessage(exchange);

            // TODO Delete the out message if the it still handled by the
            // message exchange
        } finally {
            if (exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
        }
    }

    /**
     * Process "in-out" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws MessagingException
     *             When errors occur during processing
     * 
     */
    private void processInOutAsProvider(final Exchange exchange) throws MessagingException {
        this.listener.onJBIMessage(exchange);
        // Set Fault message if no Fault or Out messages and ACTIVE Status
        if (exchange.getFault() == null && (!(((ExchangeImpl) exchange).isOutMessage()))
                && exchange.isActiveStatus()) {
            exchange.setError(new JBIException("The Provider failed to process the request"));
        }
    }

    /**
     * 
     * @param exchange
     * @throws MessagingException
     */
    private void processRobustInOnlyAsConsumer(Exchange exchange) throws MessagingException {
        try {
            this.listener.onJBIMessage(exchange);
        } finally {
            if (exchange.isActiveStatus()) {
                exchange.setDoneStatus();
            }
        }
    }

    /**
     * Process "robust-in-only" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws MessagingException
     *             When errors occur during processing
     * 
     */
    private void processRobustInOnlyAsProvider(final Exchange exchange) throws MessagingException {
        this.listener.onJBIMessage(exchange);

        // Set Status DONE if no Fault and ACTIVE Status
        if (exchange.getFault() == null && exchange.isActiveStatus()) {
            exchange.setDoneStatus();
        }
    }
}
