/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.su;

import static org.ow2.petals.component.framework.Constants.Component.AUTOGENERATED_ENDPOINT_NAME;
import static org.ow2.petals.component.framework.Constants.Component.JBI_XML;
import static org.ow2.petals.component.framework.Constants.Component.META_INF;
import static org.ow2.petals.component.framework.Constants.Component.WSDL_LOCATION;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.DeploymentException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.interceptor.MessageInterceptorManager;
import org.ow2.petals.component.framework.interceptor.description.ExtensibleElementInterceptorConfiguration;
import org.ow2.petals.component.framework.interceptor.description.InterceptorBuilder;
import org.ow2.petals.component.framework.util.Extensions;
import org.ow2.petals.component.framework.util.IDGenerator;
import org.ow2.petals.component.framework.util.ManagementMessageUtil;
import org.ow2.petals.component.framework.util.PropertiesUtil;
import org.ow2.petals.component.framework.util.WSDLHelper;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.ow2.petals.jbi.descriptor.Consumes;
import org.ow2.petals.jbi.descriptor.JBIDescriptor;
import org.ow2.petals.jbi.descriptor.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.Provides;
import org.ow2.petals.jbi.descriptor.ServiceUnitExtensibleElement;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

/**
 * A simple ServiceUnitManager. It processes service unit JBI descriptor : <br/>
 * Provides nodes : new endpoints registered in JBI container<br/> Consumes
 * nodes : link addresses to existing JBI endpoints
 * 
 * @author ofabre,alouis
 * 
 */
public class ServiceUnitManager implements javax.jbi.component.ServiceUnitManager {

    /**
     * The component's Logger
     */
    private Logger logger; // NOPMD by gblondelle

    /**
     * The service unit data handlers. Each service unit is linked to a
     * datahandler. Key is the service unit name.
     */
    private final Map<String, ServiceUnitDataHandler> serviceUnitDataHandlers;

    protected final AbstractComponent component;

    /**
     * Creates a new instance of {@link ServiceUnitManager}
     * 
     * @param component
     */
    public ServiceUnitManager(final AbstractComponent component) {
        super();
        this.component = component;
        this.logger = component.getLogger();
        this.serviceUnitDataHandlers = new HashMap<String, ServiceUnitDataHandler>();
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#deploy(java.lang.String,
     *      java.lang.String)
     */
    public String deploy(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        this.logger.log(Level.FINE, "deploy  serviceUnitName  " + serviceUnitName
                + "serviceUnitRootPath" + suRootPath);
        
        if (this.serviceUnitDataHandlers.get(serviceUnitName) == null) {
            final ServiceUnitDataHandler handler = new ServiceUnitDataHandler();
            handler.setInstallRoot(suRootPath);
            handler.setName(serviceUnitName);
            final File jbiXmlFile = new File(suRootPath + File.separator + META_INF
                    + File.separator + JBI_XML);

            JBIDescriptorBuilder builder = null;
            JBIDescriptor descriptor = null;
            try {
                builder = JBIDescriptorBuilder.getInstance();
            } catch (final JBIDescriptorException e) {
                final String msg = "Can't load jbi.xml schema";
                throw new DeploymentException(msg, e);
            }
            try {
                descriptor = builder.validateAndBuildDescriptor(jbiXmlFile.toURI(), true);
            } catch (final JBIDescriptorException e) {
                final String msg = "Bad JBI Descriptor for SU " + serviceUnitName;
                throw new DeploymentException(msg, e);
            }
            handler.setDescriptor(descriptor);

            // validate the SU extensions if validator is available
            if (this.component.getServiceUnitExtensionsValidator() != null) {
                try {
                    this.validateExtensions(descriptor);
                } catch (final PEtALSCDKException e) {
                    final String msg = "Failed to validate at least one SU extensions for SU "
                            + serviceUnitName;
                    throw new DeploymentException(msg, e);
                }
            }

            try {
                this.processConsumesNodes(handler);
            } catch (final JBIException ex) {
                final String msg = "Failed to process consumes nodes for SU " + serviceUnitName;
                throw new DeploymentException(msg, ex);
            }

            this.logger.log(Level.FINEST, "Register a data handler for SU : " + serviceUnitName);
            this.serviceUnitDataHandlers.put(serviceUnitName, handler);

            // additional deployment by component implementation
            if (this.component.getServiceUnitListener() != null) {
                try {
                    this.component.getServiceUnitListener().onSUDeployed(serviceUnitName,
                            suRootPath, descriptor);
                } catch (final Throwable e) {
                    // clean cache if any exception
                    this.logger.log(Level.FINEST, "Unregister a data handler for SU : "
                            + serviceUnitName);
                    this.serviceUnitDataHandlers.remove(serviceUnitName);

                    throw new DeploymentException(
                            "The ServiceUnit listener failed to process for SU " + serviceUnitName,
                            e);
                }
            }
        } else {
            throw new DeploymentException(
                    "A service unit with the given name is already registered : " + serviceUnitName);
        }
        final String result = ManagementMessageUtil.getComponentTaskResult(this.component
                .getContext().getComponentName(), "deploy",
                ManagementMessageUtil.TASK_RESULT_SUCCESS);
        return result;
    }

    /**
     * Retrieve the Provides node that has activated the given jbi Endpoint
     * 
     * @param serviceEndpoint
     *            the previously activated jbi Endpoint
     * @return the {@link Provides} element
     */
    public Provides getProvidesFromEndpoint(final ServiceEndpoint serviceEndpoint) {
        Provides result = null;
        Iterator<ServiceUnitDataHandler> sudh = this.serviceUnitDataHandlers.values().iterator();

        boolean find = false;
        while (sudh.hasNext() && !find) {
            ServiceUnitDataHandler handler = sudh.next();
            logger.fine("Got SU datahandler : " + handler);
            final Map<ServiceEndpoint, Provides> epJBIDescs = handler.getEpJBIDesc();
            result = epJBIDescs.get(serviceEndpoint);
            logger.fine("Got provides : " + result);
            if (result != null) {
                find = true;
            }
        }
        return result;
    }

    /**
     * Retrieve the Consumes node matching the given destination information
     * (endpoint name, service name, interface name). Search order is by
     * ascending priority : endpointName, serviceName, interfaceName. If there
     * are many consumes matching information, return the first one found.
     * 
     * @param endpointName
     *            an endpoint name, can be null
     * @param serviceName
     *            a service {@link QName}, can be null
     * @param interfaceName
     *            an interface {@link QName}, can be null
     * @return the {@link Consumes} element matching information, can be null
     */
    public Consumes getConsumesFromDestination(final String endpointName, final QName serviceName,
            final QName interfaceName) {
        Consumes result = null;
        if (endpointName != null) {
            result = this.searchConsumesMatchingEndpoint(endpointName);
        }
        if (result == null) {
            if (serviceName != null) {
                result = this.searchConsumesMatchingService(serviceName);
            }
            if (result == null) {
                if (interfaceName != null) {
                    result = this.searchConsumesMatchingInterface(interfaceName);
                }
            }
        }
        return result;
    }

    /**
     * 
     * @param endpointName
     * @return
     */
    private Consumes searchConsumesMatchingEndpoint(final String endpointName) {
        Consumes result = null;
        for (Iterator<ServiceUnitDataHandler> iter = serviceUnitDataHandlers.values().iterator(); iter
                .hasNext()
                && result == null;) {
            ServiceUnitDataHandler handler = iter.next();
            List<Consumes> consumesList = handler.getConsumesList();

            for (Iterator<Consumes> iterator = consumesList.iterator(); iterator.hasNext()
                    && result == null;) {
                Consumes consumes = iterator.next();
                String epName = consumes.getEndpointName();
                if (endpointName.equals(epName)) {
                    result = consumes;
                }
            }
        }
        return result;
    }

    /**
     * 
     * @param serviceName
     * @return
     */
    private Consumes searchConsumesMatchingService(final QName serviceName) {
        Consumes result = null;
        for (Iterator<ServiceUnitDataHandler> iter = serviceUnitDataHandlers.values().iterator(); iter
                .hasNext()
                && result == null;) {
            ServiceUnitDataHandler handler = iter.next();

            List<Consumes> consumesList = handler.getConsumesList();
            for (Iterator<Consumes> iterator = consumesList.iterator(); iterator.hasNext()
                    && result == null;) {
                Consumes consumes = iterator.next();
                QName servName = consumes.getServiceName();
                if (serviceName.equals(servName)) {
                    result = consumes;
                }
            }
        }
        return result;
    }

    /**
     * 
     * @param interfaceName
     * @return
     */
    private Consumes searchConsumesMatchingInterface(final QName interfaceName) {
        Consumes result = null;
        for (Iterator<ServiceUnitDataHandler> iter = serviceUnitDataHandlers.values().iterator(); iter
                .hasNext()
                && result == null;) {
            ServiceUnitDataHandler handler = iter.next();

            List<Consumes> consumesList = handler.getConsumesList();
            for (Iterator<Consumes> iterator = consumesList.iterator(); iterator.hasNext()
                    && result == null;) {
                Consumes consumes = iterator.next();
                QName intName = consumes.getInterfaceName();
                if (interfaceName.equals(intName)) {
                    result = consumes;
                }
            }
        }
        return result;
    }

    /**
     * Retrieve the Service description (as DOM Document) of the given jbi
     * Endpoint
     * 
     * @param serviceEndpoint
     *            the jbi Endpoint
     * @return the service description as Dom Document
     */
    public Document getServiceDescription(final ServiceEndpoint serviceEndpoint) {
        Document result = null;
        for (final Iterator<ServiceUnitDataHandler> iter = this.serviceUnitDataHandlers.values()
                .iterator(); iter.hasNext() && (result == null);) {
            final ServiceUnitDataHandler handler = iter.next();
            result = handler.getEndpointDescription(serviceEndpoint);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#init(java.lang.String,
     *      java.lang.String)
     */
    public void init(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        this.logger.log(Level.FINE, "init  serviceUnitName  " + serviceUnitName
                + "  serviceUnitRootPath " + suRootPath);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#shutDown(java.lang.String)
     */
    public void shutDown(final String serviceUnitName) throws DeploymentException {
        // stop the providing of services for the given SU, as said in the JBI
        // specifications
        this.logger.log(Level.FINE, "shutdown  serviceUnitName  " + serviceUnitName);

        try {
            final ServiceUnitDataHandler handler = this.serviceUnitDataHandlers
                    .get(serviceUnitName);
            final List<ServiceEndpoint> endpoints = this.getServiceUnitEndpoints(serviceUnitName);
            for (final ServiceEndpoint endpoint : endpoints) {
                this.component.getContext().deactivateEndpoint(endpoint);
                handler.removeEndpoint(endpoint);
            }

        } catch (final JBIException e) {
            throw new DeploymentException("Deactivate endpoint failed for SU " + serviceUnitName, e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#start(java.lang.String)
     */
    public void start(final String serviceUnitName) throws DeploymentException {
        this.logger.log(Level.FINE, "start serviceUnitName  " + serviceUnitName);

        final ServiceUnitDataHandler handler = this.serviceUnitDataHandlers.get(serviceUnitName);
        try {
            this.processProvidesNodes(handler);
        } catch (final JBIException ex) {
            final String msg = "Failed to start SU " + serviceUnitName;
            this.logger.log(Level.SEVERE, msg, ex);
            throw new DeploymentException(msg, ex);
        }
        if (this.component.getServiceUnitListener() != null) {
            try {
                this.component.getServiceUnitListener().onSUStarted(serviceUnitName);
            } catch (final PEtALSCDKException e) {
                throw new DeploymentException("The ServiceUnit listener failed to process for SU "
                        + serviceUnitName, e);
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#stop(java.lang.String)
     */
    public void stop(final String serviceUnitName) throws DeploymentException {
        this.logger.log(Level.FINE, "stop  serviceUnitName  " + serviceUnitName);

        if (this.component.getServiceUnitListener() != null) {
            try {
                this.component.getServiceUnitListener().onSUStopped(serviceUnitName);
            } catch (final PEtALSCDKException e) {
                throw new DeploymentException("The ServiceUnit listener failed to process for SU "
                        + serviceUnitName, e);
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#undeploy(java.lang.String,
     *      java.lang.String)
     */
    public String undeploy(final String serviceUnitName, final String serviceUnitRootPath)
            throws DeploymentException {
        final ServiceUnitDataHandler handler = this.serviceUnitDataHandlers.get(serviceUnitName);

        try {
            if (this.component.getServiceUnitListener() != null) {
                this.component.getServiceUnitListener().onSUUndeployed(serviceUnitName,
                        serviceUnitRootPath, handler.getDescriptor());

            }
        } catch (final PEtALSCDKException e) {
            // instead of throwing an exception, return an error message
            ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                    .getComponentName(), "undeploy", ManagementMessageUtil.TASK_RESULT_FAILED,
                    ManagementMessageUtil.MESSAGE_TYPE_ERROR, "ERROR",
                    "The ServiceUnit listener failed to process", new String[] { serviceUnitName,
                            serviceUnitRootPath }, e.getMessage());
            // throw new DeploymentException("The ServiceUnit listener failed to
            // process", e);
        } finally {
            this.logger.log(Level.FINEST, "Unregister a data handler for SU " + serviceUnitName);
            this.serviceUnitDataHandlers.remove(serviceUnitName);
        }

        return ManagementMessageUtil.getComponentTaskResult(this.component.getContext()
                .getComponentName(), "undeploy", ManagementMessageUtil.TASK_RESULT_SUCCESS);
    }

    /**
     * Search the wsdl description of the activated endpoint in the Provides
     * extensions. If no wsdl description can be found, it creates a light
     * weight Wsdl description from Provides Interface, Service and Endpoint
     * names.
     * 
     * @param installRoot
     *            the SU install root
     * @param provides
     *            the Provices node
     * @return the WSDL description as Dom Document
     */
    @SuppressWarnings("unchecked")
    protected Document getEnpointWSDLDesc(final String installRoot, final Provides provides)
            throws PEtALSCDKException {
        Document wsdlDesc = null;

        final Extensions extensions = new Extensions(provides.getExtensions());

        // Now try to replaced optionnal place holders by real external
        // property values
        String propertiesFilePath = component.getComponentConfiguration().getPropertiesFile();
        try {
            extensions.setPetalsExtensions(PropertiesUtil.fillPlaceHolder(propertiesFilePath,
                    extensions.getPetalsExtensions()));
        } catch (PEtALSCDKException e2) {
            getLogger().log(Level.WARNING, "Cannot fill place holder of SU extensions", e2);
        }

        if (extensions != null) {
            String wsdlFileLocation = extensions.getPetalsExtensions().getProperty(WSDL_LOCATION);

            if ((wsdlFileLocation != null) && (!wsdlFileLocation.trim().equals(""))) {
                boolean urlLocation = true;
                URL url = null;
                try {
                    url = new URL(wsdlFileLocation);
                } catch (final MalformedURLException e1) {
                    urlLocation = false;
                }

                if (urlLocation) {
                    try {
                        wsdlDesc = WSDLHelper.createDocumentFromWSDL(url);
                    } catch (final Exception e) {
                        throw new PEtALSCDKException("Can not retrieve WSDL document from URL : "
                                + url);
                    }
                } else {
                    // try to get the relative path
                    if (!wsdlFileLocation.startsWith(File.separator)) {
                        wsdlFileLocation = File.separator + wsdlFileLocation;
                    }

                    final File wsdlFile = new File(installRoot + wsdlFileLocation);
                    if (!wsdlFile.exists()) {
                        throw new PEtALSCDKException(
                                "Can not retrieve WSDL document from location : "
                                        + wsdlFileLocation);
                    }
                    wsdlDesc = WSDLHelper.createDocumentFromWSDL(wsdlFile);
                }
            }
        }

        // if the WSDL description is null, generate it
        if (wsdlDesc == null) {
            wsdlDesc = WSDLHelper.createLightWSDL20(provides.getInterfaceName(), provides
                    .getServiceName(), provides.getEndpointName());
        }

        return wsdlDesc;
    }

    /**
     * 
     * @param handler
     * @param consumes
     * @throws JBIException
     */
    protected void processConsumesNode(final ServiceUnitDataHandler handler, final Consumes consumes)
            throws JBIException {

        if (consumes != null) {
            handler.addConsumes(consumes);
        }

        this.processInterceptors(handler, consumes);
    }

    /**
     * Process the interceptors for a consume/provide node.
     * 
     * @param handler
     * @param consumes
     * @throws
     */
    protected void processInterceptors(final ServiceUnitDataHandler handler,
            final ServiceUnitExtensibleElement suExtensibleElement) throws JBIException {
        final Extensions extensions = new Extensions(suExtensibleElement.getExtensions());
        final ExtensibleElementInterceptorConfiguration configuration = this
                .loadInterceptorConfigurationFromSU(extensions);

        // create the manager for this extension element. Each element has a
        // linked manager with preloaded interceptors.
        if (configuration != null) {
            final MessageInterceptorManager interceptorManager = new MessageInterceptorManager(this
                    .getLogger());
            interceptorManager.init(this.component.getInterceptors(), configuration
                    .getInputInterceptors(), configuration.getOutputInterceptors(), configuration
                    .getInputResponseInterceptors(), configuration.getOutputResponseInterceptors());

            handler.addMessageInterceptorManager(suExtensibleElement, interceptorManager);
        }
    }

    /**
     * 
     * @param handler
     * @throws JBIException
     */
    protected void processConsumesNodes(final ServiceUnitDataHandler handler) throws JBIException {
        final JBIDescriptor descriptor = handler.getDescriptor();
        final List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (final Consumes consumes : consumesList) {
            this.processConsumesNode(handler, consumes);
        }
    }

    /**
     * 
     * @param extensions
     * @return an interceptor configuration or null if no interceptor has been
     *         defined in the extensions.
     */
    protected ExtensibleElementInterceptorConfiguration loadInterceptorConfigurationFromSU(
            final Extensions extensions) {
        final InterceptorBuilder builder = InterceptorBuilder.getInstance();
        return builder.loadSUInterceptor(extensions);
    }

    /**
     * Process the given provides node.
     * 
     * @param handler
     * @param provides
     * @return The resulted Service Endpoint
     * @throws JBIException
     */
    protected ServiceEndpoint processProvidesNode(final ServiceUnitDataHandler handler,
            final Provides provides) throws JBIException {

        String endpointName = provides.getEndpointName();

        if (AUTOGENERATED_ENDPOINT_NAME.equals(endpointName)) {
            endpointName = IDGenerator.getNewID();

            // update the provides object to set the real endpoint
            // then, if the SU is restarted, the real endpoint is
            // used
            provides.setEndpointName(endpointName);

            final File jbiXmlFile = new File(handler.getInstallRoot() + File.separator + META_INF
                    + File.separator + JBI_XML);

            // update the descriptor of the SU to set the real endpoint
            // then, if the SU is recovered, the real endpoint is
            // used
            // TODO: dirty fix, create a File from the JBIBuilder instead
            try {
                Document jbiDocument = JBIDescriptorBuilder.getInstance().parseJbiXml(
                        new FileInputStream(jbiXmlFile));
                // the root node can be only a 'jbi' element
                Node servicesNode = null;
                for (final Node child : XMLUtil.getNodeChildren(jbiDocument.getFirstChild())) {
                    if (child.getLocalName().equals("services")) {
                        servicesNode = child;
                        break;
                    }
                }
                if (servicesNode != null) {
                    for (final Node child : XMLUtil.getNodeChildren(servicesNode)) {
                        if (child.getLocalName().equals("provides")
                                && (XMLUtil.getAttributeValue(child, "endpoint-name")
                                        .equals(AUTOGENERATED_ENDPOINT_NAME))) {
                            final NamedNodeMap attrs = child.getAttributes();
                            final Node serviceAttr = attrs.getNamedItem("service-name");
                            if ((serviceAttr != null)
                                    && (serviceAttr.getNodeValue().endsWith(provides
                                            .getServiceName().getLocalPart()))) {
                                final Node endpointAttr = attrs.getNamedItem("endpoint-name");
                                endpointAttr.setNodeValue(provides.getEndpointName());
                                break;
                            }
                        }
                    }
                    FileOutputStream fileOutputStream = new FileOutputStream(jbiXmlFile);
                    XMLUtil.writeDocument(jbiDocument, fileOutputStream);
                }
            } catch (Exception e) {
                throw new JBIException(e);
            }
        }

        final ServiceEndpoint ep = this.component.getContext().activateEndpoint(
                provides.getServiceName(), endpointName);
        try {
            handler.addEndpointDescription(ep, this.getEnpointWSDLDesc(handler.getInstallRoot(),
                    provides));
        } catch (final JBIException e) {
            // deactive the endpoint as the start process has failed
            this.component.getContext().deactivateEndpoint(ep);
            throw e;
        }
        handler.addJBIDescription(ep, provides);
        this.processInterceptors(handler, provides);

        return ep;
    }

    /**
     * 
     * @param handler
     * @throws JBIException
     */
    protected void processProvidesNodes(final ServiceUnitDataHandler handler) throws JBIException {
        final JBIDescriptor descriptor = handler.getDescriptor();

        final List<Provides> providesList = descriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            ServiceEndpoint sep = this.processProvidesNode(handler, provides);
            this.logger.log(Level.INFO, "New Service Endpoint has been deployed : " + sep);
        }
    }

    /**
     * 
     * @param serviceUnitName
     * @return
     */
    private List<ServiceEndpoint> getServiceUnitEndpoints(final String serviceUnitName) {
        return new ArrayList<ServiceEndpoint>(this.serviceUnitDataHandlers.get(serviceUnitName)
                .getEpServiceDesc().keySet());
    }

    /**
     * Return all {@link ServiceUnitDataHandler}s as a Map
     * 
     * @return all {@link ServiceUnitDataHandler}s as a Map
     */
    public Map<String, ServiceUnitDataHandler> getServiceUnitDataHandlers() {
        return this.serviceUnitDataHandlers;
    }

    /**
     * Return the component Logger
     * 
     * @return the component Logger
     */
    public Logger getLogger() {
        return this.logger;
    }

    /**
     * Retrieve the install root of the Service Unit that activated the given
     * endpoint
     * 
     * @param ep
     *            the previously activated endpoint
     * @return the Service unit install root
     */
    public String getSUInstallRootForActivatedEp(final ServiceEndpoint ep) {
        final ServiceUnitDataHandler handler = this.getSuDataHandlerForEp(ep);
        return handler.getInstallRoot();
    }

    /**
     * Retrieve the data handler of the Service Unit that activated the given
     * endpoint.
     * 
     * @param ep
     *            the previously activated endpoint, or <code>null</code> if
     *            not found.
     * @return the {@link ServiceUnitDataHandler}
     */
    public ServiceUnitDataHandler getSuDataHandlerForEp(final ServiceEndpoint ep) {
        ServiceUnitDataHandler result = null;
        if (ep != null) {
            for (final Iterator<ServiceUnitDataHandler> iter = this.serviceUnitDataHandlers
                    .values().iterator(); iter.hasNext() && (result == null);) {
                final ServiceUnitDataHandler element = iter.next();

                for (final Iterator<ServiceEndpoint> iterator2 = element.getEpJBIDesc().keySet()
                        .iterator(); iterator2.hasNext() && (result == null);) {
                    final ServiceEndpoint element2 = iterator2.next();
                    if (ep.equals(element2)) {
                        result = element;
                    }
                }
            }
        }
        return result;
    }

    /**
     * 
     * @param extensible
     * @return
     */
    public ServiceUnitDataHandler getSuDataHandlerForConsumeProvide(
            final ServiceUnitExtensibleElement extensible) {
        ServiceUnitDataHandler result = null;

        if (extensible != null) {
            for (final Iterator<ServiceUnitDataHandler> iter = this.serviceUnitDataHandlers
                    .values().iterator(); iter.hasNext() && (result == null);) {
                final ServiceUnitDataHandler element = iter.next();

                for (final Iterator<ServiceUnitExtensibleElement> iterator2 = element
                        .getMessageInterceptorManager().keySet().iterator(); iterator2.hasNext()
                        && (result == null);) {

                    final ServiceUnitExtensibleElement element2 = iterator2.next();
                    if (extensible.equals(element2)) {
                        result = element;
                    }
                }
            }
        }

        return result;
    }

    /**
     * Validate the SU extensions.
     * 
     * @param descriptor
     *            the service unit descriptor
     * @throws PEtALSCDKException
     *             if at least an extension block is not valid. It is the role
     *             of the developer to specify which extension is not valid.
     */
    protected void validateExtensions(final JBIDescriptor descriptor) throws PEtALSCDKException {

        // validate the provides
        final List<Provides> providesList = descriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            this.component.getServiceUnitExtensionsValidator().validateProvidesExtensions(
                    provides.getExtensions());
        }

        // validate the consumes
        final List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (final Consumes consumes : consumesList) {
            this.component.getServiceUnitExtensionsValidator().validateConsumesExtensions(
                    consumes.getExtensions());
        }
    }
}
