/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * Extensions.java - Initial Developper : ofabre
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.util;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.ow2.petals.component.framework.exception.PEtALSCDKException;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Node;

import static org.ow2.petals.jbi.descriptor.JBIDescriptorConstants.NAMESPACE_PETALS_EXTENSIONS;

/**
 * An util class that parses Extensions given as constructor parameters to
 * extract Petals specific Extensions. Petals key value extensions are mapped to
 * a Properties object and can be accessed directly with the
 * getPetalsExtensions() method or getValue() methods.
 *
 * @author ofabre - EBM Websourcing
 *
 */
public class Extensions extends org.ow2.petals.jbi.descriptor.Extensions {

    public static final String PETALS_PARAMETERS_NAME = "params";

    private static final String PETALS_PARAMETER_ATTRIBUTE_NAME = "name";

    private org.ow2.petals.jbi.descriptor.Extensions extensions;

    private Properties petalsExtensions;
    
    /**
     * The PEtALS parameters
     */
    private Set<String> parameters;

    /**
     * CDK Extensions. Load extensions from jbi.xml and external properties file
     *
     * @param extensions
     *            classical jbi descriptor extensions. Must be non null
     */
    public Extensions(final org.ow2.petals.jbi.descriptor.Extensions extensions) {
        super();
        if (extensions == null) {
            throw new NullPointerException("extensions constructor parameter must not be null");
        }
        this.extensions = extensions;
        this.petalsExtensions = new Properties();
        this.parameters = new HashSet<String>();
        this.loadExtensions(extensions.getDocumentFragment());
    }

    /**
     * Searches for the Petals extension with the specified key in the Petals
     * extension list. The method returns the default value argument if the
     * property is not found.
     *
     * @param key
     *            the key of the extension
     * @param defaultValue
     *            a default value for this extension
     * @return a Petals extension as String for the specified key
     */
    public String getValue(final String key, final String defaultValue) {
        return this.petalsExtensions.getProperty(key, defaultValue);
    }

    /**
     * Searches for the Petals extension with the specified key in the Petals
     * extension list.
     *
     * @param key
     *            the key of the extension
     * @param defaultValue
     *            a default value for this extension
     * @return a Petals extension as String for the specified key or null if not
     *         found
     */
    public String getValue(final String key) {
        return this.petalsExtensions.getProperty(key);
    }

    public void addPetalsExtension(final String key, final String value) {
        this.petalsExtensions.put(key, value);
    }

    /**
     * Return all Petals extensions as {@link Properties} object
     *
     * @return Petals specific Extensions as {@link Properties} object
     */
    public Properties getPetalsExtensions() {
        return this.petalsExtensions;
    }

    /**
     * Return all extensions (Petals or not) as {@link DocumentFragment}
     */
    @Override
    public DocumentFragment getDocumentFragment() {
        return this.extensions.getDocumentFragment();
    }

    @Override
    public void setDocumentFragment(final DocumentFragment documentFragment) {
        this.extensions.setDocumentFragment(documentFragment);
    }

    /**
     * Extension nodes are nodes which names are different from those specified
     * by JBI. This method load these extensions.
     *
     * @param node
     *            an extension xml node
     * @throws PEtALSCDKException
     */
    private void loadExtensions(final Node node) {
        if(node != null) {
            final List<Node> children = XMLUtil.getNodeChildren(node);

            for (final Node child : children) {
                // Load extension information
                final String uriString = child.lookupNamespaceURI(child.getPrefix());
                // Check if the node uri is a PEtALS uri
                if ((Extensions.PETALS_PARAMETERS_NAME.equals(child.getLocalName()))
                        && (NAMESPACE_PETALS_EXTENSIONS.equals(uriString))) {
                    // get parameters includes in params tag
                    this.loadPetalsParameters(child);
                } else if (NAMESPACE_PETALS_EXTENSIONS.equals(uriString)) {
                    // get specific petals parameters
                    this.loadPetalsKeyValueExtension(child);
                }
            }
        }
    }

    public static Map<String, String> getPetalsParameters(final Node child) {
        final Map<String, String> parameters = new HashMap<String, String>();
        if ((Extensions.PETALS_PARAMETERS_NAME.equals(child.getLocalName()))
                && (NAMESPACE_PETALS_EXTENSIONS.equals(child.getNamespaceURI()))) {
            final List<Node> children = XMLUtil.getNodeChildren(child);
            for (final Node extChild : children) {
                final Node parametersNameAttribute = extChild.getAttributes()
                .getNamedItem(Extensions.PETALS_PARAMETER_ATTRIBUTE_NAME);
                if (parametersNameAttribute != null) {
                    if (XMLUtil.getTextContent(extChild) == null) {
                        parameters.put(parametersNameAttribute.getNodeValue(),
                                "");
                    } else {
                        parameters.put(parametersNameAttribute.getNodeValue(),
                                XMLUtil.getTextContent(extChild));
                    }
                }
            }
        }
        return parameters;
    }

    /**
     * 
     * @param child
     */
    private void loadPetalsParameters(final Node child) {
        Map<String, String> extensions = Extensions.getPetalsParameters(child);
        this.parameters = extensions.keySet();
        this.petalsExtensions.putAll(extensions);
    }

    /**
     * 
     * @param child
     */
    private void loadPetalsKeyValueExtension(final Node child) {
        // treat key value nodes
        if (NAMESPACE_PETALS_EXTENSIONS.equals(child.getNamespaceURI())) {
            if (XMLUtil.getTextContent(child) == null) {
                this.addPetalsExtension(child.getLocalName(), "");
            } else {
                this.addPetalsExtension(child.getLocalName(), XMLUtil
                        .getTextContent(child));
            }
        }
    }

    public void setPetalsExtensions(Properties petalsExtensions) {
        this.petalsExtensions = petalsExtensions;
    }

    public Set<String> getParameters() {
        return parameters;
    }
}
